import random
from typing import List, Iterable

from opustrainer.alignments import format_alignments, Pair
from opustrainer.modifiers import Modifier
from opustrainer.modifiers.placeholders import get_random_unicode_words


class NoiseModifier(Modifier):
    """Adds additional sentence pairs entirely made of noise during training.
    Nonsensical string on the source and on the target.

       Usage:
       ```yaml
       modifiers:
       - Noise: 0.01
         min_word_length: 2
         max_word_length: 5
         max_words: 6
        ```
    """
    min_word_length: int
    max_word_length: int
    max_words: int

    def __init__(self, probability: float, min_word_length: int=2,
                 max_word_length: int=5, max_words: int=6):
        super().__init__(probability)
        self.min_word_length = min_word_length
        self.max_word_length = max_word_length
        self.max_words = max_words

    def __call__(self, batch:List[str]) -> Iterable[str]:
        """Generates a random noise line"""
        for line in batch:
            if self.probability > random.random():
                tokens = get_random_unicode_words(self.min_word_length, self.max_word_length, self.max_words)
                fake_line = " ".join(tokens) + "\t" + " ".join(tokens)
                
                # If there's a third field in the original line, we will also
                # add alignment information to this fake line. We assume 1-1
                # mapping.
                if line.count('\t') >= 2:
                    alignments = [Pair(i, i) for i in range(len(tokens))]
                    fake_line += "\t" + format_alignments(alignments)
                
                yield fake_line
            
            # Always also yield the original line
            yield line



import random
from typing import List, Iterable

from opustrainer.modifiers import Modifier

# Use regular ASCII punctuation and the corresponding full-width (Chinese) and other analogues
PUNCT_BY_TYPE = {
    # PERIOD / FULL STOP
    'period': {
        '.', '．', '｡', '。',  # ASCII + CJK full/half-width
        '﹒', '︒',  # CJK small / presentation
        '।', '॥',  # danda, double-danda  (hi, bn, gu, kn, ml, mr, ta, te)
        '۔',  # Arabic full stop     (ar, fa, ur)
        '။',  # Myanmar stop         (my)
        '።',  # Ethiopic stop        (am)
        '։'  # Armenian stop        (hy)
    },

    # EXCLAMATION MARK
    'exclamation': {
        '!', '！',  # ASCII + CJK full-width
        '﹗', '︕',  # CJK small / presentation
        '‼', '❢', '❣',  # double & heavy
        '՜'  # Armenian exclamation (hy)
    },

    # QUESTION MARK
    'question': {
        '?', '？',  # ASCII + CJK full-width
        '﹖', '︖',  # CJK small / presentation
        '؟',  # Arabic question      (ar, fa, ur)
        ';',  # Greek question mark  (el)
        '՞',  # Armenian question    (hy)
        '፧',  # Ethiopic question    (am)
        '⁇', '⁈', '⁉'  # double / interrobang
    },

    # COMMA
    'comma': {
        ',', '，', '､',  # ASCII + CJK full/half-width
        '﹐', '︐',  # CJK small / presentation
        '、', '﹑', '︑',  # ideographic comma variants
        '،',  # Arabic comma         (ar, fa, ur)
        '၊',  # Myanmar comma        (my)
        '՝'  # Armenian comma       (hy)
    },

    # COLON
    'colon': {
        ':', '：',  # ASCII + CJK full-width
        '﹕', '︓',  # CJK small / presentation
        '፥', '፦',  # Ethiopic colons       (am)
        '·'  # Greek ano-teleia — semicolon/colon (el)
    },

    # SEMICOLON
    'semicolon': {
        ';', '；',  # ASCII + CJK full-width
        '﹔', '︔',  # CJK small / presentation
        '؛',  # Arabic semicolon     (ar, fa, ur)
        '፤'  # Ethiopic semicolon   (am)
    },
    'ellipsis': {
        '…',  # U+2026  horizontal ellipsis
        '⋯',  # U+22EF  mid-line ellipsis
        '︙',  # U+FE19  vertical CJK ellipsis
        # Don't deal with this case as it complicates handling alignments, skip lines with multiple end punctuation instead
        # '...'  # three consecutive ASCII dots
    }
}

ALL_PUNCT = {mark for _, marks in PUNCT_BY_TYPE.items() for mark in marks}


class RemoveEndPunctuationModifier(Modifier):
    """
    Removes punctuation in the end of the sentence if it matches for the source and target
    """

    def remove_punct(self, line: str) -> str:
        sections: List[str] = line.split('\t')
        src = sections[0]
        trg = sections[1]
        out_sections = [src[:-1].rstrip(), trg[:-1].rstrip()]

        if len(sections) == 3:
            aln = sections[2]
            puct_pos_src = len(src.split(' ')) - 1
            puct_pos_trg = len(trg.split(' ')) - 1
            # alignments format: 0-0 1-1 2-4 5-6
            # remove pairs that include end punctuation positions
            new_aln_parts = [pair for pair in aln.split(' ') if
                             int(pair.split('-')[0]) != puct_pos_src and int(pair.split('-')[1]) != puct_pos_trg]
            new_aln = ' '.join(new_aln_parts)
            out_sections.append(new_aln)

        return '\t'.join(out_sections)

    def ends_with_punct(self, line: str) -> bool:
        sections: List[str] = line.split('\t')
        src, trg = sections[0], sections[1]

        for kind, punct in PUNCT_BY_TYPE.items():
            # Both ends of source and target sentence should have punctuation of the same kind (e.g. `Hello. 你好。`)
            if src[-1] in punct and trg[-1] in punct:
                # Skip lines with multiple terminal punctuation marks (e.g. `Hello!?` or `Hello...`)
                for fragment in [src, trg]:
                    if len(fragment) > 1 and fragment[-2] in ALL_PUNCT:
                        return False
                    # When tokenized (e.g. `Hello . . .`)
                    if len(fragment) > 2 and fragment[-2] == ' ' and fragment[-3] in ALL_PUNCT:
                        return False
                return True

        return False

    def __call__(self, batch: List[str]) -> Iterable[str]:
        for line in batch:
            yield self.remove_punct(line) \
                if self.probability > random.random() and self.ends_with_punct(line) \
                else line



import logging
import os
import random

from multiprocessing import Queue, Process
from logging.handlers import QueueHandler, QueueListener
from typing import List, Union
from itertools import chain

from opustrainer.modifiers import Modifier


class ModifierWorker(Process):
    """Process that runs batches of sentences through a list of modifiers"""
    tasks: Queue
    results: Queue
    messages: Queue
    modifiers: List[Modifier]

    def __init__(self, tasks:Queue, results:Queue, messages:Queue, modifiers:List[Modifier], **kwargs):
        super().__init__(**kwargs)
        self.tasks = tasks
        self.results = results
        self.messages = messages
        self.modifiers = modifiers
        
    def run(self):
        handler = QueueHandler(self.messages)
        logging.getLogger().addHandler(handler)
    
        while True:
            task = self.tasks.get()

            # If task is None, this worker can stop.
            if task is None:
                break

            # A task consists of a chunk id, batch seed, and lines
            chunk, seed, batch = task

            try:
                # Set random seed for this batch, so the worker and the order
                # in which batches are processed are no longer relevant
                random.seed(seed)

                for modifier in self.modifiers:
                    batch = list(modifier(batch))

                self.results.put((chunk, batch, None))
            except Exception as exc:
                self.results.put((chunk, None, exc))
        self.results.close()


class ModifierPool:
    """Pool of ModifierWorker that exposes `map()` to run a batch of sentences
    through a predefined list of modifiers. Similar to multiprocessing.Pool
    except that the `func` argument doesn't need to be passed for each call.
    """

    """Number of worker processes in the pool"""
    workers: int

    """Modifier list each worker applies to the batches"""
    modifiers: List[Modifier]

    """Queue for submitting chunks of work to the workers"""
    tasks: Queue

    """Queue for receiving chunks from the workers"""
    results: Queue

    messages: Queue

    log_worker: QueueListener

    def __init__(self, modifiers:List[Modifier], processes:int=0):
        self.modifiers = modifiers
        self.workers = processes if processes > 0 else min(os.cpu_count() or 1, 8)

    def __enter__(self) -> 'ModifierPool':
        self.tasks = Queue()
        self.results = Queue()
        
        self.messages = Queue()
        self.log_worker = QueueListener(self.messages, *logging.getLogger().handlers, respect_handler_level=True)
        self.log_worker.start()

        self.processes = [
            ModifierWorker(self.tasks, self.results, self.messages, self.modifiers, daemon=True)
            for _ in range(self.workers)
        ]

        for process in self.processes:
            process.start()

        return self

    def __exit__(self, *args):
        # Tell workers to stop
        for _ in self.processes:
            self.tasks.put(None)
        self.tasks.close()

        # Wait for workers to close down
        for process in self.processes:
            process.join()

        self.log_worker.stop()
        self.messages.close()

    def map(self, batch:List[str], chunksize:int=0) -> List[str]:
        if chunksize > 0:
            chunks, remainder = divmod(len(batch), chunksize)
        else:
            chunks = len(self.processes)
            chunksize, remainder = divmod(len(batch), chunks)

        # Shortcut for getting the slice of batch for each chunk
        chunk_slice = lambda chunk: slice(
            chunk * chunksize,
            chunk * chunksize + (chunksize if chunk < chunks else remainder)
        )

        # Submit tasks to workers
        for chunk in range(chunks + (1 if remainder > 0 else 0)):
            self.tasks.put((chunk, random.random(), batch[chunk_slice(chunk)]))

        # Placeholder for the returned chunks, in order
        chunk_results = [[]] * (chunks + (1 if remainder > 0 else 0))

        # Retrieve results from workers
        for _ in range(chunks + (1 if remainder > 0 else 0)):
            chunk, result, exc = self.results.get()
            if exc is not None:
                raise exc
            chunk_results[chunk] = result

        # Stitch the ordered result chunks back together into a single batch
        return list(chain(*chunk_results))


class ErzatsModifierPool:
    """Same as ModifierPool, but does all the work on the main thread."""
    def __init__(self, modifiers:List[Modifier], processes:int=0):
        self.modifiers = modifiers

    def __enter__(self) -> 'ErzatsModifierPool':
        return self

    def __exit__(self, *args):
        pass

    def map(self, batch:List[str], chunksize:int=0) -> List[str]:
        if chunksize > 0:
            chunks, remainder = divmod(len(batch), chunksize)
        else:
            raise ValueError("Need a chunksize > 0")

        # Shortcut for getting the slice of batch for each chunk
        chunk_slice = lambda chunk: slice(
            chunk * chunksize,
            chunk * chunksize + (chunksize if chunk < chunks else remainder)
        )

        tasks = []

        # Submit tasks to workers
        for chunk in range(chunks + (1 if remainder > 0 else 0)):
            tasks.append((chunk, random.random(), batch[chunk_slice(chunk)]))

        # Placeholder for the returned chunks, in order
        chunk_results = [[]] * (chunks + (1 if remainder > 0 else 0))

        random_state = random.getstate()

        for chunk, seed, batch in tasks:
            # Set random seed for this batch, so the worker and the order
            # in which batches are processed are no longer relevant
            random.seed(seed)

            for modifier in self.modifiers:
                batch = list(modifier(batch))

            chunk_results[chunk] = batch

        random.setstate(random_state)
        
        # Stitch the ordered result chunks back together into a single batch
        return list(chain(*chunk_results))


def make_modifier_pool(modifiers:List[Modifier], processes:int) -> Union[ModifierPool, ErzatsModifierPool]:
    if processes == 0:
        return ErzatsModifierPool(modifiers, processes)
    else:
        return ModifierPool(modifiers, processes)



#!/usr/bin/env python3
"""A translation model trainer. It feeds marian different sets of datasets with different thresholds
for different stages of the training. Data is uncompressed and TSV formatted src\ttrg
"""
import os
import sys
import signal
import argparse
import random
import subprocess
import shlex
import time

from dataclasses import dataclass
from typing import List, Tuple, Dict, Any, Optional, Union, Type, TextIO, cast, Iterable, Iterable, TypeVar, get_type_hints, get_args, get_origin
from tempfile import TemporaryFile
from itertools import islice
from pathlib import Path

import yaml

from opustrainer.modifiers import Modifier
from opustrainer.modifiers.merge import MergeModifier
from opustrainer.modifiers.noise import NoiseModifier
from opustrainer.modifiers.prefix import PrefixModifier
from opustrainer.modifiers.punctuation import RemoveEndPunctuationModifier
from opustrainer.modifiers.surface import UpperCaseModifier, TitleCaseModifier
from opustrainer.modifiers.placeholders import PlaceholderTagModifier
from opustrainer.modifiers.typos import TypoModifier
from opustrainer.modifiers.retokenize import RetokenizeModifier
from opustrainer.modifiers.pool import make_modifier_pool
from opustrainer import logger

def ignore_sigint():
    """Used as pre-exec hook for the trainer program as to ignore ctrl-c. We'll
    deal with ctrl-c in the python program, and then be very friendly about
    stopping the trainer.
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)


# Available batch modifiers
# TODO: Import these lazy, on demand?
MODIFIERS = {
    'Merge': MergeModifier,
    'Noise': NoiseModifier,
    'UpperCase': UpperCaseModifier,
    'TitleCase': TitleCaseModifier,
    'RemoveEndPunct': RemoveEndPunctuationModifier,
    'Tags': PlaceholderTagModifier,
    'Typos': TypoModifier,
    'Prefix': PrefixModifier,
    'Retokenize': RetokenizeModifier,
}

@dataclass(frozen=True)
class Dataset:
    name: str
    files: List[str]


@dataclass(frozen=True)
class DatasetState:
    seed: int
    line: int
    epoch: int


@dataclass(frozen=True)
class Stage:
    name: str
    datasets: List[Tuple[Dataset, float]]
    until_dataset: str
    until_epoch: Optional[int]
    modifiers: Optional[List[Modifier]]


@dataclass(frozen=True)
class Curriculum:
    seed: int
    datasets: Dict[str,Dataset]
    stages: Dict[str,Stage]
    modifiers: List[Modifier]
    stages_order: List[str]

    # Optionally, check how many fields each row of the datasets should have
    # Too many should select the N first fields. Too few should drop the row.
    num_fields: Optional[int]

    def __post_init__(self):
        if len(self.stages) != len(frozenset(self.stages)):
            raise ValueError('stages can only occur once')

        if not (frozenset(self.stages) <= frozenset(self.stages.keys())):
            raise ValueError('each stage has to be defined')

    def next_stage(self, stage:Stage) -> Optional[Stage]:
        """Helper to get the next stage given the current stage."""
        index = self.stages_order.index(stage.name)
        if index + 1 < len(self.stages_order):
            return self.stages[self.stages_order[index + 1]]
        else:
            return None


@dataclass(frozen=True)
class EpochTrackerState:
    epoch: int
    line: int


@dataclass(frozen=True)
class TrainerState:
    stage: str
    random_state: Any # whatever the type is returned by random.getstate(), which I think is implementation specific.
    epoch_tracker_state: EpochTrackerState
    datasets: Dict[str,DatasetState]


class DatasetReader:
    """Repeats, shuffles and reads a dataset ad infinitum."""
    dataset: Dataset
    seed: int
    line: int
    epoch: int
    shuffle: bool
    num_fields: Optional[int]

    tmpdir: Optional[str]

    _fh: Optional[TextIO] = None
    _next_line: str

    def __init__(self, dataset:Dataset, seed:int, tmpdir:Optional[str]=None, shuffle:bool=True,
                 num_fields:Optional[int]=None):
        """
        Parameters
        ----------
        dataset : Dataset
            Description of the dataset and its files
        seed : int
            Seed number for the random number generator that shuffles the data internally
        tmpdir : str, optional
            Path to directory in which the temporary shuffled dataset is written (default is `tempfile.gettempdir()`)
        shuffle : bool
            Indicates whether shuffling should happen. Enabled by default.
        num_fields: int, optional
            Optionally specify the number of fields each line should have. Trim to the required number if they are
            more than the necessary fields, or remove lines that don't have the required number of fields.
        """
        self.dataset = dataset
        self.seed = seed
        self.tmpdir = tmpdir
        self.epoch = 0
        self.line = 0
        self.shuffle = shuffle
        self.num_fields = num_fields

    def state(self) -> DatasetState:
        return DatasetState(self.seed, self.line, self.epoch)

    def restore(self, state:DatasetState) -> 'DatasetReader':
        self.close()

        self.seed = state.seed
        self.epoch = state.epoch

        # Skip forward
        for _ in range(state.line):
            next(self)

        return self

    def close(self):
        if self._fh:
            self._fh.close()

    def _open(self):
        logger.log(f"Reading {self.dataset.name} for epoch {self.epoch}")
        # Open temporary file which will contain shuffled version of `cat self.files`
        fh = TemporaryFile(mode='w+', encoding='utf-8', dir=self.tmpdir)

        # Shuffle data to the temporary file.
        # TODO: With the reimplementation of shuffle.py, it is technically
        # feasible to just write to a named pipe (or even stdout) instead of
        # a temporary file, and let the trainer read directly from that. Not 
        # sure if that has any performance or stability benefits/drawbacks.
        subprocess.check_call([sys.executable,
            '-m', 'opustrainer.shuffle',
            *(['--temporary-directory', self.tmpdir] if self.tmpdir else []),
            *([] if self.shuffle else ['--no-shuffle']),
            str(self.seed),
            f'/dev/fd/{fh.fileno()}',
            *self.dataset.files
        ], pass_fds=(fh.fileno(),))

        # Replace open file handle with this new file
        self._fh = cast(TextIO, fh) # TODO: Not sure why TemporaryFile is an
                                    # IO[str] according to typing, but seems
                                    # to implement TextIO.
        self._fh.seek(0)
        self.line = 0

        # Buffer the first line, also asserting that we're not reading an empty file.
        try:
            self._read_line()
        except StopIteration:
            raise RuntimeError('reading from empty shuffled file')

    def _read_line(self) -> None:
        try:
            # Try to find the next non-empty line
            while True:
                self._next_line = self._fh.readline() # type: ignore # _fh can't be none.

                # Empty return, not even a line ending, means EOF
                if self._next_line == '':
                    raise StopIteration

                # We can't call the function inside the string format prior to python 3.12
                # so we will have it here instead so that we can refer to it when logging if necessary.
                original_line: str = self._next_line.rstrip('\r\n')

                # Assert that the line is well formed, meaning non of the fields is the empty string
                # If not, try to get a new line from the corpus
                fields: List[str] = original_line.split('\t')
                if any(field == '' for field in fields):
                    logger.log_once(f"[Trainer] Empty field in {self.dataset.name} line: \"{original_line}\", skipping...", loglevel="WARNING")
                    continue

                # Try to see if we have the right number of fields and remove lines
                # that don't have all the fields or remove extra fields.
                if self.num_fields is not None:
                    if len(fields) == self.num_fields:
                        return
                    elif len(fields) > self.num_fields:
                        self._next_line = '\t'.join(fields[:self.num_fields]) + '\n'
                        return
                    elif len(fields) < self.num_fields:
                        logger.log_once(f"[Trainer] Expected {self.num_fields} fields in {self.dataset.name} line: \"{original_line}\" but only got {len(fields)}, skipping...", loglevel="WARNING")
                        continue

                return
        except StopIteration:
            self._fh.close() # type: ignore # _fh can't be none.
            self.seed += 1
            self.epoch += 1

    def __iter__(self):
        return self

    def __next__(self) -> str:
        if not self._fh or self._fh.closed:
            self._open() # TODO: do we want to do this lazy? Yes, restore()
                         # might be called twice right now and shuffling is
                         # expensive.

        assert self._fh is not None
        line = self._next_line
        self.line += 1

        # Read next line into memory and test we're not at EOF
        self._read_line()

        return line


@dataclass(frozen=True)
class ShuffledFile:
    seed: int
    proc: subprocess.Popen
    file: TextIO


class AsyncDatasetReader(DatasetReader):
    _pending: Optional[ShuffledFile]

    def __init__(self, *args, **kwargs):
        self._pending = None
        super().__init__(*args, **kwargs)

    def _open_async(self, seed:int):
        # Open temporary file which will contain shuffled version of `cat self.files`
        fh = TemporaryFile(mode='w+', encoding='utf-8', dir=self.tmpdir)

        self._pending = ShuffledFile(
            seed=seed,
            file=cast(TextIO, fh),
            proc=subprocess.Popen([sys.executable,
                '-m', 'opustrainer.shuffle',
                *(['--temporary-directory', self.tmpdir] if self.tmpdir else []),
                *([] if self.shuffle else ['--no-shuffle']),
                str(seed),
                f'/dev/fd/{fh.fileno()}',
                *self.dataset.files
            ], pass_fds=(fh.fileno(),))
        )

    def _kill_async(self):
        if self._pending is None:
            return

        self._pending.proc.kill()
        self._pending.proc.wait()
        self._pending.file.close()
        self._pending = None

    def _open(self):
        logger.log(f"Reading {self.dataset.name} for epoch {self.epoch}")

        # First time self._pending is None, but all subsequent calls to _open
        # should have self._pending be set.
        if self._pending is None:
            self._open_async(self.seed)

        # Assume shuffling has started
        assert self._pending is not None
        assert self._pending.seed == self.seed

        # Wait for that to finish (hopefully it already has since it was likely
        # started last iteration)
        self._pending.proc.wait()
        assert self._pending.proc.returncode == 0

        # Swap out the current _fh for the newly prepared one
        assert self._fh is None or self._fh.closed
        self._fh = self._pending.file
        self._pending = None

        # Make sure we start reading from the start again
        self._fh.seek(0)
        self.line = 0

        # Buffer the first line, also asserting that we're not reading an empty file.
        try:
            self._read_line()
        except StopIteration:
            raise RuntimeError('reading from empty shuffled file')

        # Start shuffling next
        self._open_async(self.seed + 1)

    def restore(self, state:DatasetState) -> 'AsyncDatasetReader':
        # Note: super().restore() will call close(), which will stop any
        # running shuffling that is probably no longer relevant.
        # TODO: Once PEP 673 is available, we can remove this overload entirely.
        return cast('AsyncDatasetReader', super().restore(state))

    def close(self):
        self._kill_async()
        super().close()


class StateLoader:
    """Tool to read and write TrainerState objects to yaml. Uses unsafe yaml
    because `random.getstate()` basically returns a blob, and it is very
    particular about the data types of that structure. So we use the yaml loader
    that encodes the python data type as well (i.e. tuple vs list).
    """
    def load(self, fh:TextIO) -> TrainerState:
        ymldata = yaml.load(fh, Loader=yaml.Loader)
        if not isinstance(ymldata, dict):
            raise ValueError(f'Empty state file: {fh.name}')
        return TrainerState(
            stage=ymldata['stage'],
            random_state=ymldata['random_state'],
            epoch_tracker_state=ymldata['epoch_tracker_state'],
            datasets={
                dataset_name: DatasetState(int(seed), int(line), int(epoch))
                for dataset_name, [seed, line, epoch] in ymldata['datasets'].items()
            }
        )

    def dump(self, state:TrainerState, fh:TextIO) -> None:
        yaml.dump({
            'stage': state.stage,
            'random_state': state.random_state,
            'epoch_tracker_state': state.epoch_tracker_state,
            'datasets': {
                dataset_name: [state.seed, state.line, state.epoch] #TODO: why a tuple, why not a dict? Isn't a dict more forward compatible?
                for dataset_name, state in state.datasets.items()
            }
        }, fh, allow_unicode=True, sort_keys=False) #TODO: is safe_dump not sufficient?


class CurriculumLoaderError(ValueError):
    """Exception raised when the yaml data contains an invalid curriculum"""
    pass


def flatten(iterable):
    """Iterates recursively (depth-first) trough all items in a list with
    possibly nested lists"""
    for item in iterable:
        if isinstance(item, list):
            yield from flatten(item)
        else:
            yield item


class CurriculumV1Loader:
    def load(self, ymldata:dict, *, basepath:str='./') -> Curriculum:
        datasets = self._load_datasets(ymldata, basepath)
        stages_order = self._load_stage_order(ymldata)
        return Curriculum(
            seed=int(ymldata['seed']),
            datasets=datasets,
            stages_order=stages_order,
            stages=self._load_stages(ymldata, basepath, stages_order, datasets),
            modifiers=self._load_modifiers(ymldata, basepath),
            num_fields=int(ymldata['num_fields']) if 'num_fields' in ymldata else None
        )

    def _load_datasets(self, ymldata:dict, basepath:str) -> Dict[str,Dataset]:
        """Reads
        ```yml
        datasets:
          clean: path/to/clean.gz
        ```
        """
        return {
            name: Dataset(name, [os.path.join(basepath, filepath)])
            for name, filepath in ymldata['datasets'].items()
        }

    def _load_stage_order(self, ymldata:dict) -> List[str]:
        """Reads
        ```yaml
        stages:
          - stage1
          - stage2
        ```
        """
        return list(ymldata['stages'])

    def _load_stages(self, ymldata:dict, basepath:str, stages_order:List[str], datasets:Dict[str,Dataset]) -> Dict[str,Stage]:
        """Reads:
        ```yaml
        stagename:
          - dataset1 frac
          - dataset2 frac
          - until dataset3 epochs
        ```
        or the more verbose version
        ```yaml
        stagename:
          mix:
            - dataset1 frac
            - dataset2 frac
            - until dataset3 epochs
          modifiers:
            - Modifier: freq
        ```
        """
        return {
            stage_name: self._load_stage(ymldata, basepath, stage_name, datasets, int(ymldata['seed']))
            for stage_name in stages_order
        }

    def _load_stage(self, ymldata:dict, basepath:str, stage_name:str, available_datasets:Dict[str,Dataset], seed:int) -> Stage:
        datasets: List[Tuple[Dataset, float]] = []

        if isinstance(ymldata[stage_name], list):
            mix = ymldata[stage_name]
        else:
            mix = ymldata[stage_name].get('mix', [])

        if len(mix) == 0:
            raise CurriculumLoaderError(f"the dataset mix of stage '{stage_name}' is empty or missing its until clause")
        if not mix[-1].startswith('until '):
            raise CurriculumLoaderError(f"the last entry in {stage_name}'s dataset mix is not the until clause")

        for line in mix[:-1]:
            dataset_name, weight = line.split()
            try:
                datasets.append((available_datasets[dataset_name], float(weight)))
            except KeyError:
                raise CurriculumLoaderError(f"stage '{stage_name}' mentions unknown dataset '{dataset_name}'")
            except ValueError:
                raise CurriculumLoaderError(f"could not convert the weight '{weight}' of stage '{stage_name}' dataset '{dataset_name}' to float")

        try:
            _, until_dataset_name, max_epochs = mix[-1].split()
            until_epoch = int(max_epochs) if max_epochs != 'inf' else None
        except ValueError:
            raise CurriculumLoaderError(f"could not parse last line as `until <dataset_name> <number|'inf'>`: {mix[-1]}")
        if until_dataset_name not in {dataset.name for dataset, weight in datasets if weight > 0.0}:
            raise CurriculumLoaderError(f"until clause of stage '{stage_name}' watches dataset '{until_dataset_name}' but that dataset is not read during this stage")

        try:
            return Stage(
                name=stage_name,
                datasets=datasets,
                until_dataset=until_dataset_name,
                until_epoch=until_epoch,
                modifiers=self._load_modifiers(ymldata[stage_name], basepath) if isinstance(ymldata[stage_name], dict) and 'modifiers' in ymldata[stage_name] else None
            )
        except Exception as exc:
            raise CurriculumLoaderError(f"could not complete the parse of stage '{stage_name}': {exc!s}") from exc

    def _load_modifiers(self, ymldata:dict, basepath:str) -> List[Modifier]:
        """Reads
        ```yml
        modifiers:
          - UpperCase: 0.05
          - TitleCase: 0.05
          - Tags: 0.02
            num_tags: 6
            custom_detok_src: null
            custom_detok_trg: zh
        ```
        """
        modifiers = [
            self._load_modifier(modifier_entry, basepath)
            for modifier_entry in flatten(ymldata.get('modifiers', []))
        ]

        for modifier in modifiers:
            modifier.validate(modifiers)

        return modifiers

    def _load_modifier(self, modifier_entry:Dict[str, Any], basepath:str) -> Modifier:
        (name, probability), *config_pairs = modifier_entry.items()
        settings = {
            **dict(config_pairs),
            'probability': float(probability)
        }
        try:
            modifier = MODIFIERS[name]
        except KeyError:
            raise CurriculumLoaderError(f"unknown modifier '{name}'")
        try:
            args = self._match_modifier_signature(modifier, settings, basepath)
            return modifier(**args)
        except Exception as exc:
            raise CurriculumLoaderError(f"could not initialize modifier '{name}': {exc!s}") from exc

    def _match_modifier_signature(self, fn:Type[Modifier], settings:Dict[str, Any], basepath:str) -> Dict[str,Any]:
        """Casts `settings` values into the types expected by `fn`'s constructor. This mainly
        exists so that a modifier can define a `arg:Path` argument that is then correctly
        interpreted as a relative path."""
        hints = get_type_hints(fn.__init__)
        args = {}

        for name, value in settings.items():
            if name in hints:
                try:
                    value = self._dynamic_cast_parameter(hints[name], value, basepath)
                except Exception as exc:
                    raise CurriculumLoaderError(f"could not convert '{name}' to the correct type: {exc!s}") from exc
            args[name] = value
        
        return args

    T = TypeVar('T')

    def _dynamic_cast_parameter(self, typedef:Type[T], value:Any, basepath:str) -> Union[T, None, Path]:
        """Casts `value` into a type that matches `typedef`. Can deal with `Union`."""
        if get_origin(typedef) == Union:
            attempt_errors = ''
            for subtype in get_args(typedef):
                try:
                    return self._dynamic_cast_parameter(subtype, value, basepath)
                except Exception as err:
                    attempt_errors += f'could not cast to {subtype!r}: {err!s}\n'
            raise ValueError(attempt_errors)
        elif typedef == Path:
            return Path(basepath).joinpath(value)
        elif typedef == type(None) and value is None:
            return None
        else:
            return typedef(value)


class CurriculumLoader:
    """Reads curriculum yaml files. Wrapper that decides which reader to use
    based on a version number that may be in there.
    """

    IMPLEMENTATIONS={
        '1': CurriculumV1Loader,
    }

    def load(self, fh:Union[TextIO,str,dict], **kwargs) -> Curriculum:
        if isinstance(fh, dict):
            ymldata = fh
        else:
            ymldata = yaml.safe_load(fh)

        impl = self.IMPLEMENTATIONS[str(ymldata.get('version', '1'))]()
        return impl.load(ymldata, **kwargs)


class EpochTracker:
    """Utility to track how many epochs the reader has progressed since the
    tracker started tracking."""
    def __init__(self, reader:DatasetReader):
        self.reader = reader
        self.epoch_offset = reader.epoch
        self.line_offset = reader.line

    @property
    def epoch(self):
        epoch = self.reader.epoch - self.epoch_offset

        # ... but if the reader is behind on where it was, it hasn't completed
        # a full epoch yet.
        if self.reader.line < self.line_offset:
            epoch -= 1
        return epoch

    def restore(self, state:EpochTrackerState) -> 'EpochTracker':
        self.epoch_offset = state.epoch
        self.line_offset = state.line
        return self

    def state(self) -> EpochTrackerState:
        return EpochTrackerState(self.epoch_offset, self.line_offset)


class Trainer:
    """Writes lines to a trainer program according to the curriculum."""
    curriculum: Curriculum
    readers: Dict[str, DatasetReader]
    stage: Optional[Stage]
    epoch_tracker: EpochTracker

    # Path to write temporary shuffled files to
    tmpdir:Optional[str]
    # For debugging purposes, whether to shuffle or not
    shuffle:bool

    # Reader class to use (I.e. DatasetReader or AsyncDatasetReader)
    _reader_impl: Type[DatasetReader]

    def __init__(self, curriculum:Curriculum, *, reader:Type[DatasetReader] = DatasetReader, \
                 tmpdir:Optional[str]=None, shuffle:bool=True):
        self.curriculum = curriculum
        self.tmpdir = tmpdir
        self.shuffle = shuffle
        self._reader_impl = reader
        random.seed(self.curriculum.seed)
        first_stage_name = self.curriculum.stages_order[0]

        #TODO: make sure this doesn't do too much work in case we call
        # restore() manually anyway.
        self.restore(TrainerState(
            stage=first_stage_name,
            random_state=random.getstate(),
            epoch_tracker_state=EpochTrackerState(0, 0),
            datasets={
                dataset.name: DatasetState(seed=curriculum.seed, line=0, epoch=0)
                for dataset in curriculum.datasets.values()
            }
        ))

    def restore(self, state:TrainerState):
        random.setstate(state.random_state)
        self.stage = self.curriculum.stages[state.stage]
        self.readers = {
            dataset.name: self._reader_impl(dataset, self.curriculum.seed,
                tmpdir=self.tmpdir,
                shuffle=self.shuffle,
                num_fields=self.curriculum.num_fields
            ).restore(state.datasets[dataset.name])
            for dataset in self.curriculum.datasets.values()
        }
        self.epoch_tracker = EpochTracker(self.readers[self.stage.until_dataset]).restore(state.epoch_tracker_state)

    def state(self) -> TrainerState:
        return TrainerState(
            stage=self.stage.name if self.stage is not None else '',
            random_state=random.getstate(),
            epoch_tracker_state=self.epoch_tracker.state(),
            datasets={
                name: reader.state()
                for name, reader in self.readers.items()
            }
        )

    def close(self):
        for reader in self.readers.values():
            reader.close()
        self.readers = {}

    def next_stage(self) -> Optional[Stage]:
        """Move to the next stage. Will return this next stage or None if there is no next stage."""
        if self.stage is None:
            return None

        self.stage = self.curriculum.next_stage(self.stage)

        # If there is a next stage, also reset the epoch tracker to track the
        # `until` clause of that new stage.
        # TODO: when self.stage is None, should we delete the EpochTracker?
        if self.stage is not None:
            self.epoch_tracker = EpochTracker(self.readers[self.stage.until_dataset])

        return self.stage

    def log_stage_information(self):
        if not self.stage:
            return

        logger.log(f'Starting stage: "{self.stage.name}"')
        if self.stage.until_epoch:
            logger.log(
                f'Running until dataset "{self.stage.until_dataset}" for {self.stage.until_epoch} epoch(s)'
            )
        else:
            logger.log(f'Running until dataset "{self.stage.until_dataset}"')

        if self.curriculum.modifiers:
            logger.log("Using curriculum modifiers:")
            for modifier in self.curriculum.modifiers:
                name = type(modifier).__name__
                logger.log(f" - {name} ({modifier.probability})")

        if self.stage.modifiers:
            logger.log("Using stage modifiers:")
            for modifier in self.stage.modifiers:
                name = type(modifier).__name__
                logger.log(f" - {name} ({modifier.probability})")

    def run(self, *, batch_size:int=100, chunk_size:int=16, processes:int=0) -> Iterable[List[str]]:
        """Yield batches, moving through the stages of training as datasets are consumed."""
        while self.stage is not None:
            self.log_stage_information()

            # Stage level modifiers take precedence over global modifiers,
            # but you can combine them yourself using YAML references.
            if self.stage.modifiers is not None:
                modifiers = self.stage.modifiers
            else:
                modifiers = self.curriculum.modifiers

            with make_modifier_pool(modifiers, processes) as pool:
                while self.stage.until_epoch is None or self.epoch_tracker.epoch < self.stage.until_epoch:
                    batch: List[str] = []

                    # Read from each dataset according to its weight in this stage
                    # (They will reshuffle and repeat if necessary)
                    for dataset, weight in self.stage.datasets:
                        batch.extend(
                            line.rstrip('\r\n') for line in
                            islice(self.readers[dataset.name], 0, int(batch_size * weight))
                        )

                    # Apply any modifiers to random lines in the batch, or sentence
                    # (Multiple modifiers can be applied to the same line)
                    batch = pool.map(batch, chunk_size)

                    if self.shuffle:
                        random.shuffle(batch)

                    # Tell anyone whose listening that something interesting happened
                    # TODO: Yield something useful, e.g. progress.
                    yield [line + '\n' for line in batch]

            # Move onto next stage. May be `None`, which would end this generator
            self.next_stage()
        logger.log("Finished the last stage")

class StateTracker:
    """Wraps around the trainer.run() call to restore and dump state its."""
    path: str
    loader: StateLoader
    dump: bool
    restore: bool

    def __init__(self, path:str, *, loader:StateLoader=StateLoader(), restore:bool=True, dump:bool=True, timeout=60):
        """
        Parameters
        --–-------
        path : str
            Path to state file
        loader : type, optional
            Loader class for encoding/decoding the state file
        restore : bool, optional
            Whether to restore the state if the state file currently exists (default is True)
        dump : bool, optional
            Whether to dump the state to the file after training (default is True)
        timeout : int, optional
            Minimum number of seconds between state dumps
        """
        self.path = path
        self.loader = loader
        self.dump = dump
        self.restore = restore
        self.timeout = timeout
        self._last_dump = 0

    def _restore(self, trainer:Trainer):
        with open(self.path, 'r', encoding='utf-8') as fh:
            return trainer.restore(self.loader.load(fh))

    def _dump(self, trainer:Trainer):
        new_statefile = f"{self.path}.new"
        with open(new_statefile, 'w', encoding='utf-8') as fh:
            self.loader.dump(trainer.state(), fh)
        os.rename(new_statefile, self.path)
        self._last_dump = time.monotonic()

    def run(self, trainer:Trainer, *args, **kwargs):
        if self.restore and os.path.exists(self.path):
            self._restore(trainer)

        try:
            for batch in trainer.run(*args, **kwargs):
                # TODO: Replace this with something that listens to Marian, and
                # writes the state to disk after marian performed validation.
                if self.dump and time.monotonic() - self._last_dump > self.timeout:
                    self._dump(trainer)
                yield batch
        finally:
            # Dump on clean exit as well as on exception.
            if self.dump:
                self._dump(trainer)


def print_state(state:TrainerState) -> None:
    logger.log(f"At stage {state.stage}")
    for name, reader in state.datasets.items():
        logger.log(f"Dataset {name}: overall epochs {reader.epoch: 3d}.{reader.line:010d}")


def parse_args(args:List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Feeds marian tsv data for training.", allow_abbrev=False)
    parser.add_argument("--config", '-c', required=True, type=str, help='YML configuration input.')
    parser.add_argument("--state", '-s', type=str, help='YML state file, defaults to ${CONFIG}.state.')
    parser.add_argument("--sync", action="store_true", help="Do not shuffle async")
    parser.add_argument("--temporary-directory", '-T', default=None, type=str, help='Temporary dir, used for shuffling and tracking state')
    parser.add_argument("--do-not-resume", '-d', action="store_true", help='Do not resume from the previous training state')
    parser.add_argument("--no-shuffle", '-n', action="store_false", help='Do not shuffle, for debugging', dest="shuffle")
    parser.add_argument("--batch-size", '-b', type=int, default=100, help='Batch size')
    parser.add_argument("--chunk-size", '-B', type=int, default=16, help='Chunk size of batches fed to modifiers')
    parser.add_argument("--workers", '-j', type=int, default=os.cpu_count() or 1, help='Number of workers')
    parser.add_argument("--log-level", type=str, default="INFO", help="Set log level. Available levels: DEBUG, INFO, WARNING, ERROR, CRITICAL. Default is INFO")
    parser.add_argument("--log-file", '-l', type=str, default=None, help="Target location for logging. Always logs to stderr and optionally to a file.")
    parser.add_argument("trainer", type=str, nargs=argparse.REMAINDER, help="Trainer program that gets fed the input. If empty it is read from config.")
    return parser.parse_args(args)


def main() -> None:
    args = parse_args(sys.argv[1:])

    logger.setup_logger(args.log_file, args.log_level)

    with open(args.config, 'r', encoding='utf-8') as fh:
        config = yaml.safe_load(fh)

    curriculum = CurriculumLoader().load(config, basepath=os.path.dirname(args.config))

    # Quick cheap check that all files exist before we begin training
    for dataset in curriculum.datasets.values():
        missing_files = {file for file in dataset.files if not os.path.exists(file)}
        if missing_files:
            raise ValueError(f"Dataset '{dataset.name}' is missing files: {missing_files}")

    trainer = Trainer(curriculum,
        reader=DatasetReader if args.sync else AsyncDatasetReader,
        tmpdir=args.temporary_directory,
        shuffle=args.shuffle)

    state_tracker = StateTracker(args.state or f'{args.config}.state', restore=not args.do_not_resume)

    # Make trainer listen to `kill -SIGUSR1 $PID` to print dataset progress
    signal.signal(signal.SIGUSR1, lambda signum, handler: print_state(trainer.state()))

    model_trainer = subprocess.Popen(
        args.trainer or shlex.split(config['trainer']),
        stdin=subprocess.PIPE,
        encoding="utf-8",
        preexec_fn=ignore_sigint) # ignore_sigint makes marian ignore Ctrl-C. We'll stop it from here.

    assert model_trainer.stdin is not None

    # TODO: This logic looks complicated, should be able to do this simpler. Three scenarios:
    #   1. ctrl-c is pressed and trainer is told this is the end of the training data
    #   2. ctrl-c is pressed and trainer has much training data in its buffers, ctrl-c needs to be
    #      pressed again to tell trainer to really terminate. Just closing its stdin and waiting for
    #      it to notice takes too long
    #   3. trainer decides it has read enough and will train no longer. This is the BrokenPipeError
    #      scenario. We don't need to deal with multiple levels of terminating the trainer because
    #      the trainer is already dead at this point.
    try:
        try:
            for batch in state_tracker.run(trainer, batch_size=args.batch_size, chunk_size=args.chunk_size, processes=args.workers):
                model_trainer.stdin.writelines(batch)
        except KeyboardInterrupt:
            logger.log("Ctrl-c pressed, stopping training")

        # Levels of waiting for the trainer. This is reached either because we ran out of batches
        # or because ctrl-c was pressed. Pressing ctrl-c more advances to next level of aggressiveness.
        for stage in ['exit', 'terminate', 'kill']:
            try:
                if stage == 'exit':
                    model_trainer.stdin.close()
                elif stage == 'terminate':
                    model_trainer.terminate()
                else:
                    model_trainer.kill()
                
                logger.log(f"waiting for trainer to {stage}. Press ctrl-c to be more aggressive")
                sys.exit(model_trainer.wait()) # blocking
            except KeyboardInterrupt:
                continue
    except BrokenPipeError:
        # BrokenPipeError is thrown by writelines() or close() and indicates that the child trainer
        # process is no more. We can safely retrieve its return code and exit with that, it should
        # not block at this point.
        logger.log("trainer stopped reading input")
        sys.exit(model_trainer.wait())


if __name__ == '__main__':
    main()



import random
import unittest
from opustrainer.modifiers.retokenize import RetokenizeModifier


VOCAB = 'contrib/test-data/vocab.zhen.spm'


def first(it):
  return next(iter(it))


class TestRetokenizer(unittest.TestCase):
  maxDiff = None

  def test_identity(self):
    """The default tokenizer and detokenizer together should have no impact on
    the output."""
    tokenizer = RetokenizeModifier(
      src=dict(detokenize='spaces', tokenize=f'spaces'),
      trg=dict(detokenize='spaces', tokenize=f'spaces'))

    test_pair = [
      'This is a sentence split by spaces .',
      'Is this a sentence that is split by spaces ?',
      '0-1 1-0 2-2 3-3 4-6 5-7 6-8 7-9',
    ]

    self.assertEqual(first(tokenizer(['\t'.join(test_pair)])).split('\t'), test_pair)

  def test_multiple_spaces(self):
    """Spaces tokenizer should have no impact on the output, but it will 
    normalize spaces due to how it is implemented."""
    tokenizer = RetokenizeModifier(
      src=dict(detokenize='spaces', tokenize=f'spaces'),
      trg=dict(detokenize='spaces', tokenize=f'spaces'))

    test_pair = [
      'This is a    sentence split by spaces .',
      'Is this a sentence    that is split by spaces ?',
      '0-1 1-0 2-2 3-3 4-6 5-7 6-8 7-9',
    ]

    ref_pair = [
      'This is a sentence split by spaces .',
      'Is this a sentence that is split by spaces ?',
      '0-1 1-0 2-2 3-3 4-6 5-7 6-8 7-9',
    ]

    self.assertEqual(first(tokenizer(['\t'.join(test_pair)])).split('\t'), ref_pair)

  def test_retokenize(self):
    tokenizer = RetokenizeModifier(
      src=dict(detokenize='moses:en', tokenize=f'spm:{VOCAB}'),
      trg=dict(detokenize='moses:zh', tokenize=f'spm:{VOCAB}'))

    out = tokenizer(['\t'.join([
      'This is a simple test statement 🤣 .',
      #^0   ^1 ^2 ^3    ^4   ^5        ^6 ^7
      '这 是 一个 简单 的 测试 语 句 🤣 。',
      #^0 ^1 ^2  ^3   ^4 ^5   ^6 ^7 ^8 ^9
      '0-0 1-1 2-2 3-3 3-4 4-5 5-6 5-7 6-8 7-9',
    ])])
    self.assertEqual(first(out), '\t'.join([
      'This is a simple test statement 🤣.',
      #[This][ is][ a][ simple][ test][ statement][ ][] [] [] [🤣][.]
      #^0    ^1   ^2  ^3       ^4     ^5          ^6 ^7 ^8 ^9 ^10 ^11 
      '这是一个简单的测试语句 🤣 。',
      #[这][是][一][个][简][单][的][测][试][语][句] [ ] []  []  []  [🤣][ 。]
      #^0  ^1  ^2  ^3 ^4  ^5  ^6  ^7  ^8  ^9  ^10 ^11 ^12 ^13 ^14 ^15  ^16
      '0-0 1-1 2-2 2-3 3-4 3-5 3-6 4-7 4-8 5-9 5-10 10-15 11-16',
      # 0-0 [This]      [这]    0-0
      # 1-1 [is]        [是]    1-1
      # 2-2 [a]         [一个]  2-2 2-3
      # 3-3 [simple]    [简单]  3-4 3-5
      # 3-4 [simple]    [的]    3-6
      # 4-5 [test]      [测试]  4-7 4-8
      # 5-6 [statement] [语]    5-9
      # 5-7 [statement] [句]    5-10 (6-11)
      # 6-8 [🤣]        [🤣]   (7-12 8-13 9-14) 10-15
      # 7-9 [.]         [。]    11-16
    ]))

  def test_retokenize_icu(self):
    tokenizer = RetokenizeModifier(
      src=dict(detokenize='icu:en', tokenize=f'spm:{VOCAB}'),
      trg=dict(detokenize='icu:zh', tokenize=f'spm:{VOCAB}'))

    out = tokenizer(['\t'.join([
      'This ▁ is ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣 .',
      #^0   ^1^2 ^3^4^5^6     ^7^8   ^9^10       ^11^12^13
      '这 是 一个 简单 的 测试 语 句 ▁ 🤣 ▁ 。',
      #^0 ^1 ^2  ^3  ^4 ^5  ^6 ^7^8 ^9^10^11
      '0-0 2-1 4-2 6-3 6-4 8-5 10-6 10-7 12-9 13-11',
    ])])

    self.assertEqual(first(out), '\t'.join([
      'This is a simple test statement 🤣.',
      #[This][ is][ a][ simple][ test][ statement][ ][] [] [] [🤣][.]
      #^0    ^1   ^2  ^3       ^4     ^5          ^6 ^7 ^8 ^9 ^10 ^11
      '这是一个简单的测试语句 🤣 。',
      #[这][是][一][个][简][单][的][测][试][语][句] [ ] []  []  []  [🤣][ 。]
      #^0  ^1  ^2  ^3 ^4  ^5  ^6  ^7  ^8  ^9  ^10 ^11 ^12 ^13 ^14 ^15  ^16
      '0-0 1-1 2-2 2-3 3-4 3-5 3-6 4-7 4-8 5-9 5-10 10-15 11-16',
      # 0-0 [This]      [这]    0-0
      # 1-1 [is]        [是]    1-1
      # 2-2 [a]         [一个]  2-2 2-3
      # 3-3 [simple]    [简单]  3-4 3-5
      # 3-4 [simple]    [的]    3-6
      # 4-5 [test]      [测试]  4-7 4-8
      # 5-6 [statement] [语]    5-9
      # 5-7 [statement] [句]    5-10 (6-11)
      # 6-8 [🤣]        [🤣]   (7-12 8-13 9-14) 10-15
      # 7-9 [.]         [。]    11-16
    ]))




import unittest

from opustrainer.tokenizers import make_tokenizer, make_detokenizer

@unittest.skip("requires installing pyicu")
class TestTokenizers(unittest.TestCase):

  def test_tokenize_detokenize_icu_en(self):
    """
    Tests lossless text reconstruction by the ICU tokenizer for English.
    Requires installation with the steps specified in https://pypi.org/project/PyICU/
    """
    tokenizer = make_tokenizer('icu:en')
    detokenizer = make_detokenizer('icu:en')
    text = '“This is,” a simple test statement 🤣.'

    tokens, _ = tokenizer.tokenize(text)
    detokenized, _ = detokenizer.detokenize(tokens)

    self.assertEqual(text, detokenized)
    self.assertEqual("“ This ▁ is , ” ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣.", " ".join(tokens))


  def test_tokenize_detokenize_icu_zh(self):
    """
    Tests lossless text reconstruction by the ICU tokenizer for Chinese.
    Requires installation with the steps specified in https://pypi.org/project/PyICU/
    """
    tokenizer = make_tokenizer('icu:zh')
    detokenizer = make_detokenizer('icu:zh')
    text = '这是一个简单的测试语句 🤣 。'

    tokens, _ = tokenizer.tokenize(text)
    detokenized, _ = detokenizer.detokenize(tokens)

    self.assertEqual(text, detokenized)
    self.assertEqual("这 是 一个 简单 的 测试 语 句 ▁ 🤣▁ 。", " ".join(tokens))



import random
import unittest
from contextlib import closing

from opustrainer.trainer import Trainer, CurriculumLoader
from opustrainer.modifiers.typos import TypoModifier


INPUT_DATA = [
	("This is a sentence.", "Dit is een zin."),
	("Why are we doing this?", "Waarom zijn we dit aan het doen?"),
	("8876683648", "8876683648"),
	("---", "---"),
]


def first(it):
  return next(iter(it))


class TestTypos(unittest.TestCase):
	def test_typos(self):
		"""Test modifier only alters column 0"""
		random.seed(1)
		modifier = TypoModifier(1.0)
		output = [line.split('\t') for line in modifier(["\t".join(line) for line in INPUT_DATA])]
		# Assert first column is modified
		self.assertEqual([row[0] for row in output],[
			'Thisis a sentence.',
			'Why are we doing this?',
			'88766863648',
			'---'])
		# Assert second column is untouched.
		self.assertEqual([row[1] for row in output], [row[1] for row in INPUT_DATA])

	def test_typos_specific(self):
		"""Test each modifier is applied only once per sentence"""
		random.seed(1)
		modifier = TypoModifier(1.0, extra_char=1.0, random_space=1.0)
		output = [line.split('\t') for line in modifier(["\t".join(line) for line in INPUT_DATA])]
		self.assertEqual([row[0] for row in output], [
			'This is  a senftence.',
			'Why ad re we doing this?',
			'887 66836478',
			'-- -'])

	def test_trainer_with_typos(self):
		"""Test Typos configuration being picked up by the trainer"""
		config = {
			'datasets': {
				'clean': 'contrib/test-data/clean'
			},
			'stages': [
				'start'
			],
			'start': [
				'clean 1.0',
				'until clean 1',
			],
			'modifiers': [
				{
					'Typos': 1.0,
					'extra_char': 1.0,
					'random_space': 1.0,
				}
			],
			'seed': 1
		}
		
		curriculum = CurriculumLoader().load(config)

		# Reference batches (trainer runs without resuming)
		with closing(Trainer(curriculum)) as trainer:
			batches = list(trainer.run())


		self.assertEqual(batches[0][:5], [
			' cklean700\n',
			'c klean688\n',
			'cles an220\n',
			'clesan 281\n',
			'c lean17109\n']
		)

	def test_regression_40(self):
		random.seed(1)
		disable_all = {
			name: 0.0
			for name in TypoModifier.modifiers.keys()
		}
		line = '٦\t6'
		modifier = TypoModifier(1.0, **{**disable_all, 'extra_char':1.0})
		modified = next(iter(modifier([line])))
		self.assertNotEqual(modified, line) # test it changed
		self.assertRegex(modified, r'^..\t.$') # test it has 1 extra char

	def test_zero_prob(self):
		"""Test probability parameter (if 0, no typos)"""
		random.seed(1)
		modifier = TypoModifier(0.0, extra_char=1.0, random_space=1.0)
		output = [line.split('\t') for line in modifier(["\t".join(line) for line in INPUT_DATA])]
		self.assertEqual([row[0] for row in output], [row[0] for row in INPUT_DATA])



from io import TextIOWrapper
import logging
from sys import stderr, version_info
from typing import List, Dict, TextIO, Union, Optional
from functools import lru_cache

def _getLevelNamesMapping() -> Dict[str,int]:
    '''getLevelNamesMapping only available in python 3.11+'''
    if 'getLevelNamesMapping' in logging.__dict__:
        return logging.getLevelNamesMapping()
    else:
        return {'CRITICAL': 50,
                'FATAL': 50,
                'ERROR': 40,
                'WARN': 30,
                'WARNING': 30,
                'INFO': 20,
                'DEBUG': 10,
                'NOTSET': 0}

@lru_cache(None)
def get_log_level(name: str) -> int:
    """Incredibly, i can't find a function that will do this conversion, other
    than setLevel, but setLevel doesn't work for calling the different log level logs."""
    if name.upper() in _getLevelNamesMapping():
        return _getLevelNamesMapping()[name.upper()]
    else:
        logging.log(logging.WARNING, f"unknown log level level used: {name} assuming warning...")
        return logging.WARNING

def log(msg: str, loglevel: str = "INFO", **kwargs) -> None:
    level = get_log_level(loglevel)
    logging.log(level, msg, **kwargs)


@lru_cache(None)
def log_once(msg: str, loglevel: str = "INFO", **kwargs) -> None:
    """A wrapper to log, to make sure that we only print things once"""
    log(msg, loglevel, **kwargs)


def setup_logger(outputfilename: Optional[str] = None, loglevel: str = "INFO", disable_stderr: bool=False) -> None:
    """Sets up the logger with the necessary settings. Outputs to both file and stderr"""
    loggingformat = '[%(asctime)s] [Trainer] [%(levelname)s] %(message)s'
    handlers: List[Union[logging.StreamHandler[TextIO], logging.StreamHandler[TextIOWrapper]]] = []
    # disable_stderr is to be used only when testing the logger
    # When testing the logger directly, we don't want to write to stderr, because in order to read
    # our stderr output, we have to use redirect_stderr, which however makes all other tests spit
    # as it interferes with unittest' own redirect_stderr. How nice.
    # This happens even when assertLogs context capture is used.
    if not disable_stderr:
        handlers.append(logging.StreamHandler(stream=stderr))
    if outputfilename is not None:
        handlers.append(logging.FileHandler(filename=outputfilename))
    
    # Python 3.9 introduced an encoding argument
    if version_info[:2] >= (3,9):
        kwargs = {'encoding': 'utf-8'}
    else:
        kwargs = {}
    # This is the only logger we'd ever use. However during testing, due to the context, logger can't be recreated,
    # even if it has already been shutdown. This is why we use force=True to force recreation of logger so we can
    # properly run our tests. Not the best solution, not sure if it's not prone to race conditions, but it is
    # at the very least safe to use for the actual software running
    logging.basicConfig(handlers=handlers, level=get_log_level(loglevel), format=loggingformat,
                         datefmt='%Y-%m-%d %H:%M:%S', force=True, **kwargs) # type: ignore



import random
import re
from typing import Dict, Tuple, List, Iterable

import typo

from opustrainer.modifiers import Modifier


# Monkey patch typo to not try to find the nearest digit for
# non-digit numerals.
# See https://github.com/hplt-project/OpusTrainer/issues/40
# See https://github.com/ranvijaykumar/typo/issues/3
from  typo.keyboardlayouts import en_default
_get_random_neighbor = en_default.get_random_neighbor

def skip_non_digit_decimals(char, seed=None):
    if char.isdecimal() and char not in en_default.NEIGHBORINGNUMPADDIGITS:
        return char
    return _get_random_neighbor(char, seed)

typo.keyboardlayouts.en_default.get_random_neighbor = skip_non_digit_decimals


def add_random_space_with_alignment(
    string: str, alignment_str: str, include_edges=False
) -> Tuple[str, str]:
    """Add a space to a random point in the string"""
    possible_insertions = [m.start() for m in re.finditer(r"\S", string)]

    if include_edges:
        possible_insertions += [len(string)]  # also at end
    else:
        possible_insertions = possible_insertions[1:]  # not at beginning

    if not possible_insertions:
        return string, alignment_str

    index = random.choice(possible_insertions)
    augmented_string = string[:index] + " " + string[index:]

    # Find positions of space-like segments that would not cause de-alignment
    space_spans = list(re.finditer(r"^\s*|\s+|$", string[: index + 1]))
    keep_alignment_indices = [m.end() for m in space_spans]
    word_index = len(keep_alignment_indices) - 2

    augmented_alignment_str = alignment_str
    if index not in keep_alignment_indices:
        assert len(augmented_string.split()) - len(string.split()) == 1
        augmented_alignment = []

        for align in alignment_str.split():
            src, trg = [int(index) for index in align.split("-", maxsplit=1)]
            if src <= word_index:
                augmented_alignment.append(f"{src}-{trg}")
            if src >= word_index:
                augmented_alignment.append(f"{src+1}-{trg}")

        augmented_alignment_str = " ".join(augmented_alignment)
    return augmented_string, augmented_alignment_str


def skip_random_space_with_alignment(
    string: str, alignment_str: str, include_edges=False
) -> Tuple[str, str]:
    possible_removals = re.finditer(r"(?<!^)\s(?!$)", string)

    if include_edges:
        possible_removals = re.finditer(r"\s", string)
    possible_removals = [m.start() for m in possible_removals]

    if not possible_removals:
        return string, alignment_str

    index = random.choice(possible_removals)
    augmented_string = string[:index] + string[index + 1 :]

    word_spans = list(re.finditer(r"(?<=\S)\s\S", string))
    word_index = None
    for i, match in enumerate(word_spans):
        if index == match.start():
            word_index = i
            break

    augmented_alignment_str = alignment_str
    if word_index is not None:
        augmented_alignment = []
        for align in alignment_str.split():
            src, trg = [int(index) for index in align.split("-", maxsplit=1)]
            if src <= word_index:
                augmented_alignment.append(f"{src}-{trg}")
            else:
                augmented_alignment.append(f"{src-1}-{trg}")
        augmented_alignment_str = " ".join(augmented_alignment)

    return augmented_string, augmented_alignment_str


def missing_char_with_alignment(string: str, alignment_str: str):
    possible_removals = [m.start() for m in re.finditer(r"\S", string)]

    # Do not remove only possible char
    if len(possible_removals) <= 1:
        return string, alignment_str

    index = random.choice(possible_removals)
    augmented_string = string[:index] + string[index + 1 :]

    word_spans = list(re.finditer(r"\S+", string))

    # Find word_index if a single-char word was removed
    word_index = None
    for i, match in enumerate(word_spans):
        start, end = match.span()
        if index > end:
            continue
        if start <= index:
            if end - start == 1:
                word_index = i
            break

    augmented_alignment_string = alignment_str
    if word_index is not None:
        augmented_alignment = []
        # Reassign alignment to a random-neighbour
        neighbours = []
        if word_index != 0:
            neighbours.append(-1)
        if word_index != len(word_spans):
            neighbours.append(0)
        r = random.choice(neighbours)

        for align in alignment_str.split():
            src, trg = [int(index) for index in align.split("-", maxsplit=1)]
            if src == word_index:
                augmented_alignment.append(f"{src+r}-{trg}")
            else:
                offset = 0 if src < word_index else -1
                augmented_alignment.append(f"{src+offset}-{trg}")
        augmented_alignment_string = " ".join(augmented_alignment)

    return augmented_string, augmented_alignment_string

class TypoModifier(Modifier):
    # modifier name, and probability it is applied on a considered
    # sentence. Each modifier can either be applied once or not at all
    # for a considered sentence. The default probability for each is 10%.
    modifiers = {
        'char_swap':     0.1,
        'missing_char':  0.1,
        'extra_char':    0.1,
        'nearby_char':   0.1,
        'similar_char':  0.1,
        'skipped_space': 0.1,
        'random_space':  0.1,
        'repeated_char': 0.1,
        'unichar':       0.1,
    }

    column: int

    probabilities: Dict[str,float]

    def __init__(self, probability:float, **probabilities:float):
        """
        Apply typo modifiers to the input. If no specific typo modifiers are
        mentioned, it will default to applying them all with a 0.1 probability
        each. If modifiers are mentioned in the configuration, only the
        modifiers mentioned will be used. All probabilities have to be in the
        0.0 .. 1.0 range.

        args:
            probability: float
                probability a line will be modified

            char_swap: float
                Swaps two random consecutive word characters in the string.

            missing_char: float
                Skips a random word character in the string.

            extra_char: float
                Adds an extra, keyboard-neighbor, letter next to a random word character.

            nearby_char: float
                Replaces a random word character with keyboard-neighbor letter.

            similar_char: float
                Replaces a random word character with another visually similar character.

            skipped_space: float
                Skips a random space from the string.

            random_space: float
                Adds a random space in the string.

            repeated_char: float
                Repeats a random word character.

            unichar: float
                Replaces a random consecutive repeated letter with a single letter.
        """
        super().__init__(probability)

        for mod, mod_prob in probabilities.items():
            if mod not in self.modifiers:
                raise ValueError(f'Unknown typo modifier: {mod}')
            if mod_prob < 0.0 or mod_prob > 1.0:
                raise ValueError(f'Typo modifier {mod} has a probability out of the 0.0..1.0 range')

        self.probabilities = probabilities or self.modifiers

    def __call__(self, batch: List[str]) -> Iterable[str]:
        for line in batch:
            yield self.apply(line) if self.probability > random.random() else line

    def apply(self, line:str) -> str:
        fields = line.split("\t")

        # TODO: The StrErrer constructor calls random.seed(None), which isn't
        # great for reproducibility. Not sure whether getrandbits() is a good
        # workaround though.
        data = typo.StrErrer(fields[0], seed=random.getrandbits(32))

        has_alignment_info = len(fields) > 2

        for modifier, probability in self.probabilities.items():
            if probability > random.random():
                # Introducing spaces with alignment information is a problem.
                if has_alignment_info and modifier in (
                    "random_space",
                    "skipped_space",
                    "missing_char",
                ):
                    if modifier == "random_space":
                        m_func = add_random_space_with_alignment
                    elif modifier == "skipped_space":
                        m_func = skip_random_space_with_alignment
                    else:  # modifier == "missing_char":
                        m_func = missing_char_with_alignment

                    data.result, fields[2] = m_func(data.result, fields[2])
                else:
                    wordcount = len(data.result.split())
                    getattr(data, modifier)()
                    assert len(data.result.split()) == wordcount or not has_alignment_info, f'Modifier {modifier} changed the word count while alignment info was not updated'


        fields[0] = data.result

        return "\t".join(fields)



import unittest
import tempfile

from opustrainer import logger

class TestLogger(unittest.TestCase):
    '''Tests the logger using end-to-end config.'''
    def test_log_once(self):
        '''Tests the log_once functionality'''
        with tempfile.NamedTemporaryFile(suffix='.log', prefix="logger") as tmpfile:
            logger.setup_logger(outputfilename=tmpfile.name, disable_stderr=True)
            logger.log("Test message")
            logger.log_once("Once message")
            logger.log_once("Once message")
            logger.log("Test message")
            logger.log_once("Once message")
            logger.log_once("Once message2")
            logger.log_once("Once message2")
            logger.log("Final message")
            logger.logging.shutdown()
            # Now read
            tmpfile.seek(0)
            line1 = tmpfile.readline().decode('utf-8').strip().split(' [Trainer] ')[1]
            self.assertEqual(line1,"[INFO] Test message")
            line2 = tmpfile.readline().decode('utf-8').strip().split(' [Trainer] ')[1]
            self.assertEqual(line2,"[INFO] Once message")
            line3 = tmpfile.readline().decode('utf-8').strip().split(' [Trainer] ')[1]
            self.assertEqual(line3,"[INFO] Test message")
            line4 = tmpfile.readline().decode('utf-8').strip().split(' [Trainer] ')[1]
            self.assertEqual(line4,"[INFO] Once message2")
            line5 = tmpfile.readline().decode('utf-8').strip().split(' [Trainer] ')[1]
            self.assertEqual(line5,"[INFO] Final message")



import random
import unittest
from opustrainer.modifiers.prefix import PrefixModifier


def first(it):
  return next(iter(it))


class TestPrefix(unittest.TestCase):
  '''Tests several test cases for prefixes'''
  def test_few(self):
    random.seed(1)
    modifier = PrefixModifier(1, 2, 6)
    # Test a line with some bogus alignments
    line1 = "How are you doing today good sir?\tКак я караш днеска добри ми господине?\t1-2 2-3"
    expected1 = "__start__ караш днеска __end__ How are you doing today good sir?\tКак я караш днеска добри ми господине?\t1-2 2-3"
    self.assertEqual(first(modifier([line1])), expected1)

    # Different execution different result
    expected2 = "__start__ я караш днеска добри ми __end__ How are you doing today good sir?\tКак я караш днеска добри ми господине?\t1-2 2-3"
    self.assertEqual(first(modifier([line1])), expected2)

    # Different execution different result. This time not triggered
    expected3 = "__start__ Как я караш __end__ How are you doing today good sir?\tКак я караш днеска добри ми господине?\t1-2 2-3"
    self.assertEqual(first(modifier([line1])), expected3)
    
    # Should return the same string as the word is too short. Also tests parsing without alignment.
    line2 = "Hello\tHola"
    self.assertEqual(first(modifier([line2])), line2)




from abc import ABC, abstractmethod
from typing import NamedTuple, Dict, List, Iterable, Tuple, Optional, Any, Protocol


# List of tokens/words according to some tokenization scheme
TokenList = List[str] # todo: bytes?

# List of string offsets that accompanies a `TokenList` telling you where each
# token is found in the original string.
TokenSpanList = List[slice]

# Mapping from `old token index` => [new token indices].
TokenMapping = List[List[int]]


class Tokenizer(Protocol):
    """Turns a string into a list of tokens and slices. Each token has a slice
    at the same offset that describes where in `text` that token is found."""
    def tokenize(self, text:str) -> Tuple[TokenList,TokenSpanList]:
        ...


class Detokenizer(Protocol):
    """Turns a list of tokens into a string. The list of slices returned tells
    you where each of the input tokens is found in the detokenized string."""
    def detokenize(self, tokens:TokenList) -> Tuple[str, TokenSpanList]:
        ...


class Pair(NamedTuple):
    """Alignment pair between source and target token indices"""
    src:int
    trg:int


class SentencePair(NamedTuple):
    """Semantic representation of a single line from a data source."""
    src: TokenList
    trg: TokenList

    # alignments is an empty list if alignment data is available in the dataset
    # but there are no aligned tokens in this pair. It is None if this dataset
    # does not have alignment info.
    alignments: Optional[List[Pair]]


class Modifier(ABC):
    """Line modifier"""
    probability: float

    def __init__(self, probability:float, **kwargs:Dict[str,Any]):
        self.probability = probability

    def validate(self, context:List['Modifier']) -> None:
        """Opportunity for the modifier to see where in the modifier list it is
        placed and flag any issues to the logger. E.g. if you place a modifier that
        inserts special tokens before an UpperCase modifier, the latter might
        modify those special tokens as well. Here you can shout about that.
        """
        pass

    @abstractmethod
    def __call__(self, batch: List[str]) -> Iterable[str]:
        pass



#!/usr/bin/env python3
'''Tests the available functionality'''
import os
import tempfile
import unittest

from typing import IO, Type
from collections import Counter
from contextlib import closing
from textwrap import dedent
from io import StringIO
from itertools import chain

import yaml

from opustrainer.trainer import Curriculum, CurriculumLoaderError, Dataset, DatasetReader, AsyncDatasetReader, CurriculumLoader, Trainer, StateTracker, Stage
from opustrainer.logger import log_once

TEST_FILE: str

def setUpModule():
	global TEST_FILE
	fd, TEST_FILE = tempfile.mkstemp(text=True)
	
	with open(fd, 'w') as fh:
		for n in range(1000):
			fh.write(f'line{n}\n')


def tearDownModule():
	os.unlink(TEST_FILE)


class TestDatasetReader(unittest.TestCase):
	testset: IO[str]

	reader: Type[DatasetReader] = DatasetReader

	def test_repeating_read(self):
		"""Test whether when we read 3000 lines from a 1000 lines dataset we do
		inded read each line 3 times.
		"""
		# Read 3000 lines
		with closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader:
			counter = Counter(line for _, line in zip(range(3000), reader))

		# We should have 1000 unique lines (all lines in TEST_FILE)
		self.assertEqual(len(counter), 1000)
		# And each line should be read exactly 3 times
		self.assertEqual(set(counter[key] for key in counter.keys()), {3})
		# ideally in a different order than previous read.

	def test_shuffled_read(self):
		"""That that when we read 2000 lines from a 1000 line testfile, we read
		each line twice, but we do read them in a different order.
		"""
		# Read 3000 lines
		with closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader:
			lines1 = [line for _, line in zip(range(1000), reader)]
			lines2 = [line for _, line in zip(range(1000), reader)]
		#	We should have read the same lines
		self.assertEqual(frozenset(lines1), frozenset(lines2))
		# but in a different order
		self.assertNotEqual(lines1, lines2)

	def test_offsets(self):
		"""Test whether `epoch` and `line` properties of a DatasetReader are
		counting properly.
		"""
		with closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader:
			for _ in zip(range(500), reader):
				pass

			self.assertEqual(reader.epoch, 0)
			self.assertEqual(reader.line, 500)

			for _ in zip(range(1250), reader):
				pass

			self.assertEqual(reader.epoch, 1)
			self.assertEqual(reader.line, 750)

	def test_resume_offset(self):
		"""Test whether resuming from a DatasetReader with the same testfile and the
		same seed does indeed yield the entire dataset with no duplicates or
		omissions.
		"""
		counter = Counter()

		with closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader1:
			counter.update(line for _, line in zip(range(250), reader1))
			state = reader1.state()

		with closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader2:
			reader2.restore(state)
			counter.update(line for _, line in zip(range(750), reader2))

		# We should have read 250 + 750 lines exactly
		self.assertEqual(len(counter), 1000)
		# and they should be all read only once.
		self.assertEqual(set(counter[key] for key in counter.keys()), {1})

	def test_resume_order(self):
		"""Test whether when we resume reading a DatasetReader, the order is the
		same as if we'd read everything from the initial (non-resumed) reader.
		"""
		counter = Counter()

		with closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader1, \
			closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader2:
			lines1 = [line for _, line in zip(range(250), reader1)]
			reader2.restore(reader1.state())
			lines1.extend(line for _, line in zip(range(750), reader2))

		with closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader1, \
			closing(self.reader(Dataset('test', [TEST_FILE]), seed=1234)) as reader2:
			lines2 = [line for _, line in zip(range(500), reader1)]
			reader2.restore(reader1.state())
			lines2.extend(line for _, line in zip(range(500), reader2))

		# Both reads should have the same amount and unique sentences
		self.assertEqual(len(lines1), len(lines2))
		self.assertEqual(set(lines1), set(lines2))
		# They also should have the same order
		self.assertEqual(lines1, lines2)


class TestAsyncDatasetReader(TestDatasetReader):
	"""Run all the same tests, but on the async reader that shuffles in advance."""
	reader = AsyncDatasetReader


class TestTrainer(unittest.TestCase):
	def test_resume(self):
		"""End-to-end test for resuming training where we test that a resumed
		trainer continues with the same sentences in the same order as the original
		trainer would have, if it were to continue after dumping its state.
		"""
		config = {
			'datasets': {
				'clean': 'contrib/test-data/clean',
				'medium': 'contrib/test-data/medium',
				'dirty': 'contrib/test-data/dirty'
			},
			'stages': [
				'start',
				'mid'
			],
			'start': [
				'clean 0.8',
				'medium 0.2',
				'dirty 0',
				'until clean 1'
			],
			'mid': [
				'clean 0.6',
				'medium 0.3',
				'dirty 0.1',
				'until medium 1',
			],
			'seed': 1111
		}
		
		curriculum = CurriculumLoader().load(config)

		# Reference batches (trainer runs without resuming)
		with closing(Trainer(curriculum)) as trainer_ref:
			batches_ref = list(trainer_ref.run())

		# State tracker (using tmpdir to make sure the file does not exist)
		with tempfile.TemporaryDirectory() as tmpdir:
			state_tracker = StateTracker(os.path.join(tmpdir, 'state_file'))

			# Train on trainer1
			with closing(Trainer(curriculum)) as trainer1:
				batches = [batch for _, batch in zip(range(10), state_tracker.run(trainer1))]

			# Resume on trainer2
			with closing(Trainer(curriculum)) as trainer2:
				batches.extend(state_tracker.run(trainer2))
			
		self.assertEqual(batches, batches_ref)

	def test_deterministic_parallel(self):
		"""End-to-end test to confirm that training with 2 workers or with 4 workers
		should yield the same training data going to the trainer.
		"""
		config = {
			'datasets': {
				'clean': 'contrib/test-data/clean',
				'medium': 'contrib/test-data/medium',
				'dirty': 'contrib/test-data/dirty'
			},
			'stages': [
				'start',
				'mid'
			],
			'start': [
				'clean 0.8',
				'medium 0.2',
				'dirty 0',
				'until clean 1'
			],
			'mid': [
				'clean 0.6',
				'medium 0.3',
				'dirty 0.1',
				'until medium 1',
			],
			'modifiers': [
				{'UpperCase': 0.25}
			],
			'seed': 1111
		}
		
		curriculum = CurriculumLoader().load(config)

		# Run with one worker and default chunk size
		with closing(Trainer(curriculum)) as trainer:
			batches_linear = list(trainer.run(processes=1))

		# Run with three workers, and default batch size
		with closing(Trainer(curriculum)) as trainer:
			batches_parallel = list(trainer.run(processes=4))

		self.assertEqual(batches_linear, batches_parallel)


class TestCurriculumLoader(unittest.TestCase):
	def test_simple(self):
		"""Test loading of a minimal configuration"""
		config = {
			'datasets': {
				'clean': 'contrib/test-data/clean',
			},
			'stages': [
				'start'
			],
			'start': [
				'clean 1.0',
				'until clean 5'
			],
			'seed': 1111,
			'modifiers': [
				{'UpperCase': 1.0}
			]
		}
		curriculum = CurriculumLoader().load(config)
		self.assertIsInstance(curriculum, Curriculum)
		self.assertEqual(curriculum.datasets, {
			'clean': Dataset(name='clean', files=['./contrib/test-data/clean'])
		})
		self.assertEqual(curriculum.stages, {
			'start': Stage(
				name='start',
				datasets=[
					(Dataset('clean', ['./contrib/test-data/clean']), 1.0)
				],
				until_dataset='clean',
				until_epoch=5,
				modifiers=None)
		})
		self.assertEqual(curriculum.seed, 1111)
		self.assertEqual(len(curriculum.modifiers), 1)

	def test_no_until(self):
		"""Test that omitting the until clause raises an error"""
		config = {
			'datasets': {
				'clean': 'contrib/test-data/clean',
			},
			'stages': [
				'start'
			],
			'start': [
				'clean 1.0'
			],
			'seed': 1
		}
		with self.assertRaisesRegex(CurriculumLoaderError, 'until clause'):
			CurriculumLoader().load(config)
	
	def test_undefined_dataset(self):
		"""Test that mentioning a wrong dataset raises an error"""
		config = {
			'datasets': {
				'clean': 'contrib/test-data/clean',
			},
			'stages': [
				'start'
			],
			'start': [
				'cleany 1.0',
				'until cleany 1'
			],
			'seed': 1
		}
		with self.assertRaisesRegex(CurriculumLoaderError, 'unknown dataset'):
			CurriculumLoader().load(config)

	def test_undefined_modifier(self):
		"""Test that mentioning an unknown modifier raises an error"""
		config = {
			'datasets': {
				'clean': 'contrib/test-data/clean',
			},
			'stages': [
				'start'
			],
			'start': [
				'clean 1.0',
				'until clean 1'
			],
			'modifiers': [
				{'NonExistingModifier': 1.0}
			],
			'seed': 1
		}
		with self.assertRaisesRegex(CurriculumLoaderError, 'unknown modifier'):
			CurriculumLoader().load(config)

	def test_extended_stage_configuration(self):
		"""Test the extended stage configuration"""
		config = {
			'datasets': {
				'clean': 'contrib/test-data/clean',
			},
			'stages': [
				'start'
			],
			'start': {
				'mix': [
					'clean 1.0',
					'until clean 1'
				],
				'modifiers': [
					{'UpperCase': 1.0}
				],
			},
			'seed': 1
		}
		curriculum = CurriculumLoader().load(config)
		self.assertEqual(len(curriculum.stages['start'].datasets), 1)
		self.assertEqual(curriculum.stages['start'].until_dataset, 'clean')
		self.assertTrue(curriculum.stages['start'].modifiers is not None and len(curriculum.stages['start'].modifiers) == 1)

	def test_combined_stage_configuration(self):
		"""Test that in the extended stage configuration modifier lists are
		flattened which allows us to use YAML references more easily."""

		# Writing this in YAML directly to reflect how this is intended to be useful
		config_yaml = dedent("""
			datasets:
				clean: contrib/test-data/clean

			stages:
			- start

			modifiers: &global_modifiers
			- TitleCase: 0.5
			
			start:
				mix: 
				- clean 1.0
				- until clean 1
				modifiers:
				- UpperCase: 0.5
				- *global_modifiers
			
			seed: 1
		""").replace("\t", "  ")

		config = yaml.safe_load(StringIO(config_yaml))

		curriculum = CurriculumLoader().load(config)
		self.assertEqual([modifier.__class__.__name__ for modifier in curriculum.stages['start'].modifiers or []], ['UpperCaseModifier', 'TitleCaseModifier'])

	def test_modifier_error_line_context(self):
		"""Test that when a modifier fails, we get context information about the line that failed"""
		with tempfile.NamedTemporaryFile('w', encoding='utf-8') as fd:
			fd.write('This is a test\tDas ist ein Test\t0-0 1-1 2-2 3-3\n')
			fd.write('Hello world\tHallo Welt\t0-0 1-2\n') # 2 is out-of-bounds
			fd.flush()

			config = {
				'datasets': {
					'clean': fd.name,
				},
				'stages': [
					'start'
				],
				'start': [
					'clean 1.0',
					'until clean 1'
				],
				'modifiers': [
					{'Tags': 1.0}
				],
				'seed': 1
			}
			curriculum = CurriculumLoader().load(config)

			trainer = Trainer(curriculum)
			
			with self.assertLogs(level='WARNING') as logger_ctx:
				output = list(chain.from_iterable(trainer.run(batch_size=1)))
				# Assert we skipped the line
				self.assertEqual(len(output), 1)
				# Assert that we got the general error message
				self.assertRegex(logger_ctx.output[0], r'Skipping line because of exception:')
				# Assert that we got the specific error as well
				self.assertRegex(logger_ctx.output[0], r'ValueError\(\'Out-of-bound alignment pairs\'\)')

	def test_num_fields(self):
		"""Tests the num field limiter"""
		with tempfile.NamedTemporaryFile('w', encoding='utf-8') as fd:
			fd.write('This is a test\tDas ist ein Test\t0-0 1-1 2-2 3-3\n')
			fd.write('This is a test\tDas ist ein Test\t0-0 1-1 2-2 3-3\textra fields\n')
			fd.write('Hello world\tHallo Welt\n') # Not enough fields
			fd.flush()

			config = {
				'datasets': {
					'clean': fd.name,
				},
				'stages': [
					'start'
				],
				'start': [
					'clean 1.0',
					'until clean 1'
				],
				'seed': 1,
				'num_fields': 3
			}
			curriculum = CurriculumLoader().load(config)

			for reader in [DatasetReader, AsyncDatasetReader]:
				with self.subTest(reader=reader.__name__), \
				 self.assertLogs(level='WARNING') as logger_ctx, \
				 closing(Trainer(curriculum, reader=reader)) as trainer:
					# Reset the log_once cache
					log_once.cache_clear()

					output = list(chain.from_iterable(trainer.run(batch_size=1)))
					# Assert we skipped the line
					self.assertEqual(len(output), 2)
					# Assert we properly cropped the lines and left the others unchanged
					self.assertEqual(output, [
						'This is a test\tDas ist ein Test\t0-0 1-1 2-2 3-3\n',
						'This is a test\tDas ist ein Test\t0-0 1-1 2-2 3-3\n'
					])
					# Assert that we got an error message for one line
					self.assertRegex(logger_ctx.output[0], r'\[Trainer\] Expected 3 fields in clean line:')



import subprocess
import sys
import tempfile
import unittest
from pathlib import Path
from typing import List


class TestEndToEnd(unittest.TestCase):
    '''Tests the pipeline end-to-end. Aimed to to test consistent output.'''

    def __clean_states(self):
        """Remove state files"""
        basepath = Path("contrib")
        for state in basepath.glob("*.yml.state"):
            state.unlink(missing_ok=True)

    def setUp(self) -> None:
        self.__clean_states()

    def tearDown(self) -> None:
        self.__clean_states()

    def assertEndToEnd(self, args: List[str], path:str, *, returncode:int=0):
        process = subprocess.run(
            [sys.executable, '-m', 'opustrainer', *args],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8")

        self.assertEqual(process.returncode, returncode, msg=process.stderr)

        # Useful to generate the data, hehe
        if False and path.endswith('.expected.out'):
            with open(path, 'w', encoding='utf-8') as fh:
                fh.write(process.stdout)

        with open(path, 'r', encoding='utf-8') as fh:
            reference = fh.read()

        self.assertEqual(process.stdout, reference)

    def test_full_enzh(self):
        self.assertEndToEnd(
            ['-c', 'contrib/test_enzh_config.yml', '-d', '--sync'],
            'contrib/test-data/test_enzh_config.expected.out')

    def test_full_zhen(self):
        self.assertEndToEnd(
            ['-c', 'contrib/test_zhen_config.yml', '-d', '--sync'],
            'contrib/test-data/test_zhen_config.expected.out')

    def test_prefix_augment(self):
        self.assertEndToEnd(
            ['-c', 'contrib/test_zhen_prefix_config.yml', '-d', '--sync'],
            'contrib/test-data/test_zhen_config_prefix.expected.out')

    def test_no_shuffle(self):
        """Confirms that when an empty training procedure is used with
        the `--no-shuffle` option, it reproduces its input."""
        for mode in [[], ['--sync']]:
            with self.subTest(mode=mode):
                self.assertEndToEnd(
                    [
                        '-c', 'contrib/test_enzh_config_plain.yml',
                        '-d',
                        *mode,
                        '--no-shuffle',
                        '--batch-size', '1' # batch-size << dataset size otherwise we overproduce
                    ],
                    'contrib/test-data/clean.enzh.10')

    def test_advanced_config(self):
        self.assertEndToEnd(
            ['-c', 'contrib/test_enzh_tags_advanced_config.yml', '-d', '-n'],
            'contrib/test-data/test_enzh_tags_advanced_config.expected.out')

    def test_stage_config(self):
        self.assertEndToEnd(
            ['-c', 'contrib/test_enzh_tags_stage_config.yml', '-d', '-n'],
            'contrib/test-data/test_enzh_tags_stage_config.expected.out')

    def test_log_file_and_stderr(self):
        """Test that log messages go to stdout and to the logfile. Note that the
        log file contains 10 iterations because the dataset is only 10 lines but
        the batch size is 100. Termination conditions are only checked at the
        end of the batch. TODO: this might be a bug, or at least unexpected."""
        with tempfile.NamedTemporaryFile(suffix='.log', prefix="opustrainer", mode='w+', encoding="utf-8") as tmpfile:
            process = subprocess.run(
                [sys.executable, '-m', 'opustrainer', '-c', 'contrib/test_enzh_config_plain.yml', '-d', '-l', tmpfile.name],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
                encoding="utf-8")
            self.assertEqual(process.returncode, 0)
            
            # Check that the stderr output is the same the output written to the file
            tmpfile.seek(0)
            self.assertEqual(process.stderr, tmpfile.read())
            
            # Check if the log is the same as the reference log. This is more complicated as we have a time field
            with open('contrib/test-data/test_enzh_config_plain_expected.log', 'r', encoding='utf-8') as reflog:
                # Loglist has one extra `\n` compared to reference list, due to stderr flushing an extra empty line?
                loglist = process.stderr.splitlines(keepends=True)
                reference = reflog.readlines()
                # Strip the time field and test
                remove_timestamp = lambda line: line.split('[Trainer]', maxsplit=1)[1]
                self.assertEqual(
                    [remove_timestamp(line) for line in loglist],
                    [remove_timestamp(line) for line in reference])



from doctest import Example
import enum
import random
import unittest

from opustrainer.modifiers.noise import NoiseModifier

class TestMerge(unittest.TestCase):
  def setUp(self):
    random.seed(1)

    # Set up examples
    self.example = [
      '429 运输 中队 ( 429 野牛) , 使用 CC - 177	429 Transport Squadron (429 Bison Squadron) - Flying the CC-177	0-0 1-1 2-2 3-3 4-3 5-4 5-5 7-5 8-5 9-6 8-7 9-8 10-9',
      "微生物 检验 与 食品 安全 控制 .	Food Poisoning and Food Hygiene.	3-0 0-1 1-1 2-1 2-2 3-3 4-3 5-4 6-4"
    ]*10

    # Remove column with alignment info
    self.example_noalign = ["\t".join(a.split('\t')[:-1]) for a in self.example]

    # With 20% prob this is triggered 3 times we check one of the matches. We expect new length to be 23
    self.num_nine_noise = "쑥맜\t쑥맜"
    self.num_nine_noise_align = "쑥맜\t쑥맜\t0-0"

  def test_noise(self):
    noiser = NoiseModifier(0.2)
    noised = list(noiser(self.example_noalign))
    self.assertEqual(noised[9], self.num_nine_noise)
    self.assertEqual(len(noised), 23)
    
  def test_noise_align(self):
    noiser = NoiseModifier(0.2)
    noised = list(noiser(self.example))
    self.assertEqual(noised[9], self.num_nine_noise_align)
    self.assertEqual(len(noised), 23)



import random
from typing import List, Iterable, Optional
from itertools import accumulate, chain

from opustrainer.modifiers import Modifier
from opustrainer.alignments import format_alignments, parse_alignments, Pair
from opustrainer.tokenizers import SpaceDetokenizer, SpaceTokenizer


def merge_sents(inputs: List[str]) -> str:
    """Merges n sentences together, fixing up their alignments"""
    rows = [line.split('\t') for line in inputs]

    input_tokens = (
        [row[0].split() for row in rows], # src
        [row[1].split() for row in rows], # trg
    )

    # src and trg sentences, merged
    output_cols = [" ".join(chain(*side)) for side in input_tokens]

    # If all rows have alignment info, add a merged alignment column
    if all(len(row) > 2 for row in rows):
        # Alignments as Pairs per input sentence
        alignments = [parse_alignments(row[2]) for row in rows]

        # Offsets for where alignments per input sentence should start
        offsets = tuple(
            list(accumulate((len(sentence) for sentence in side), initial=0))
            for side in input_tokens
        )

        # New alignment pairs that have been offset to account for the sentences
        # that would precede it.
        joined_alignments = [
            Pair(p.src + offsets[0][i], p.trg + offsets[1][i])
            for i, sentence_pairs in enumerate(alignments)
            for p in sentence_pairs
        ]

        output_cols.append(format_alignments(joined_alignments))

    return "\t".join(output_cols)


class MergeModifier(Modifier):
    """Randomly merges up to n lines into one
    
        Usage:
       ```yaml
       modifiers:
       - Merge: 0.01
         min_lines: 2
         max_lines: 4
        ```
    """
    min_lines: int
    max_lines: int

    def __init__(self, probability: float, min_lines: int=2, max_lines: int=4):
        super().__init__(probability)
        self.min_lines = min_lines
        self.max_lines = max_lines

    def __call__(self, batch:List[str]) -> Iterable[str]:
        i = 0
        while i < len(batch):
            if self.probability > random.random():
                merge_size = random.randint(self.min_lines, self.max_lines)
                yield merge_sents(batch[i:i+merge_size])
                i += merge_size
            else:
                yield batch[i]
                i += 1



#!/usr/bin/env python3
import os
import unittest

import yaml

from opustrainer.trainer import parse_args
from opustrainer.trainer import CurriculumLoader


class TestArgumentParser(unittest.TestCase):
	def test_overlapping_marian_args(self):
		"""Assert that you can pass arguments that exist both in trainer and marian
		to each respectively."""
		parsed = parse_args(['--config', 'trainer.yml', 'marian', '--config', 'marian.yml'])
		expected = {
			'config': 'trainer.yml',
			'trainer': ['marian', '--config', 'marian.yml']
		}
		# The `{**vars(parsed), **expected}` bit makes it so any entries in `parsed`
		# that are not mentioned in `expected` are not tested. Once upon a time
		# `assertDictContainsSubset` existed, but that's deprecated now and can't do
		# any `|` or `<=` operators here because we're targeting Python 3.8 :sad:
		self.assertEqual({**vars(parsed), **expected}, vars(parsed))

	def test_marian_log_args(self):
		"""Assert that you can pass `--log` to trainer even when argparse will find
		this confusing."""
		parsed = parse_args(['--config', 'trainer.yml', '--log-file', 'trainer.log', 'marian', '--log', 'marian.log'])
		expected = {
			'log_file': 'trainer.log',
			'config': 'trainer.yml',
			'trainer': ['marian', '--log', 'marian.log']
		}
		self.assertEqual({**vars(parsed), **expected}, vars(parsed))

	def test_config_loader(self):
		config_path = os.path.join(
			os.path.dirname(os.path.abspath(__file__)),
			os.pardir,
			'contrib',
			'test_full_config.yml')
		with open(config_path, 'r', encoding='utf-8') as fh:
			config = yaml.safe_load(fh)

		curriculum = CurriculumLoader().load(config, basepath=os.path.dirname(config_path))

		assert curriculum is not None




from doctest import Example
import random
import unittest
from opustrainer.alignments import parse_alignments

from opustrainer.modifiers.merge import MergeModifier, merge_sents

class TestMerge(unittest.TestCase):
  def setUp(self):
    random.seed(1)

    # Set up examples
    self.example = [
      '429 运输 中队 ( 429 野牛) , 使用 CC - 177	429 Transport Squadron (429 Bison Squadron) - Flying the CC-177	0-0 1-1 2-2 3-3 4-3 5-4 5-5 7-5 8-5 9-6 8-7 9-8 10-9',
      "微生物 检验 与 食品 安全 控制 .	Food Poisoning and Food Hygiene.	3-0 0-1 1-1 2-1 2-2 3-3 4-3 5-4 6-4"
    ]*10

    # Remove the third column with alignments
    self.example_noalign = ["\t".join(a.split('\t')[:-1]) for a in self.example]

    # counts
    self.psn_cnt = " ".join(self.example).count('Poisoning') # 10
    self.num_cnt = " ".join(self.example).count('429') # 40 because it appears once in src and trg

  def test_merge_sents(self):
    merged = merge_sents(self.example_noalign[0:3])
    expected = '429 运输 中队 ( 429 野牛) , 使用 CC - 177 微生物 检验 与 食品 安全 控制 . 429 运输 中队 ( 429 野牛) , 使用 CC - 177\t429 Transport Squadron (429 Bison Squadron) - Flying the CC-177 Food Poisoning and Food Hygiene. 429 Transport Squadron (429 Bison Squadron) - Flying the CC-177'
    self.assertEqual(merged, expected)

    # Expected based on counts 
    lensrc = sum([len(a.split('\t')[0].split()) for a in self.example_noalign[0:3]])
    lentrg = sum([len(a.split('\t')[1].split()) for a in self.example_noalign[0:3]])

    lenmrgsrc = len(merged.split('\t')[0].split())
    lenmrgtrg = len(merged.split('\t')[1].split())
    self.assertEqual(lensrc, lenmrgsrc)
    self.assertEqual(lentrg, lenmrgtrg)

  def test_merge_sents_with_alignments(self):
    merged = merge_sents(self.example[0:3])
    expected = '429 运输 中队 ( 429 野牛) , 使用 CC - 177 微生物 检验 与 食品 安全 控制 . 429 运输 中队 ( 429 野牛) , 使用 CC - 177\t429 Transport Squadron (429 Bison Squadron) - Flying the CC-177 Food Poisoning and Food Hygiene. 429 Transport Squadron (429 Bison Squadron) - Flying the CC-177\t0-0 1-1 2-2 3-3 4-3 5-4 5-5 7-5 8-5 9-6 8-7 9-8 10-9 14-10 11-11 12-11 13-11 13-12 14-13 15-13 16-14 17-14 18-15 19-16 20-17 21-18 22-18 23-19 23-20 25-20 26-20 27-21 26-22 27-23 28-24'
    self.assertEqual(merged, expected)

    # Validate the alignments against the text itself (i.e. no out-of-bound tokens)
    src, trg, alignments = merged.split('\t')
    pairs = parse_alignments(alignments, src.split(), trg.split())

    # Assert all pairs are unique, and not just duplicates of the input
    self.assertEqual(len(pairs), len(set(pairs)))

    # Assert that all pairs are present
    example_pair_cnt = sum(
      len(parse_alignments(example.split('\t')[2]))
      for example in self.example[0:3]
    )
    self.assertEqual(example_pair_cnt, len(pairs))

  def test_merge_modifier_full(self):
    merger = MergeModifier(0.8)
    merged = list(merger(self.example_noalign))
    
    psn_cnt = " ".join(merged).count('Poisoning')
    num_cnt = " ".join(merged).count('429')

    self.assertNotEqual(len(merged), len(self.example_noalign)) # Assert it being activated
    self.assertEqual(self.psn_cnt, psn_cnt)
    self.assertEqual(self.num_cnt, num_cnt)

  def test_merge_modifier_full_with_alignments(self):
    merger = MergeModifier(0.8)
    merged = list(merger(self.example))
    
    psn_cnt = " ".join(merged).count('Poisoning')
    num_cnt = " ".join(merged).count('429')

    self.assertNotEqual(len(merged), len(self.example)) # Assert it being activated
    self.assertEqual(self.psn_cnt, psn_cnt)
    self.assertEqual(self.num_cnt, num_cnt)
  
 



import unittest

from parameterized import parameterized

from opustrainer.modifiers.punctuation import RemoveEndPunctuationModifier


class TestMerge(unittest.TestCase):

    @parameterized.expand([
        ('.', '。'),
        ('?', '？'),
        ('!', '!'),
        (',', '、'),
        (':', '：'),
        (';', '﹔'),
        ('…', '︙'),
    ])
    def test_remove_end_punct(self, src_punct, trg_punct):
        ex = '\t'.join([
            f'This ▁ is ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣 {src_punct}',
            f'这 是 一个 简单 的 测试 语 句 ▁ 🤣 ▁ {trg_punct}',
            '0-0 2-1 4-2 6-3 6-4 8-5 10-6 10-7 12-9 13-11',
        ])
        mod = RemoveEndPunctuationModifier(1.0)
        out = list(mod([ex]))[0]
        src, trg, aln = out.split('\t')
        self.assertEqual(src, 'This ▁ is ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣')
        self.assertEqual(trg, '这 是 一个 简单 的 测试 语 句 ▁ 🤣 ▁')
        self.assertEqual(aln, '0-0 2-1 4-2 6-3 6-4 8-5 10-6 10-7 12-9')

    def test_remove_end_punct_untokenized(self):
        ex = '\t'.join([
            f'This is a simple test statement 🤣.',
            f'这是一个简单的测试语句🤣。',
        ])
        mod = RemoveEndPunctuationModifier(1.0)
        out = list(mod([ex]))[0]
        src, trg = out.split('\t')
        self.assertEqual(src, 'This is a simple test statement 🤣')
        self.assertEqual(trg, '这是一个简单的测试语句🤣')

    def test_remove_end_punct_no_alignment(self):
        ex = '\t'.join([
            'This ▁ is ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣 .',
            '这 是 一个 简单 的 测试 语 句 ▁ 🤣 ▁ 。'
        ])
        mod = RemoveEndPunctuationModifier(1.0)
        out = list(mod([ex]))[0]
        src, trg = out.split('\t')
        self.assertEqual(src, 'This ▁ is ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣')
        self.assertEqual(trg, '这 是 一个 简单 的 测试 语 句 ▁ 🤣 ▁')

    def test_do_not_remove_end_punct_if_different(self):
        ex = '\t'.join([
            'This ▁ is ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣 ▁ ?',
            '这 是 一个 简单 的 测试 语 句 ▁ 🤣 ▁ 。',
            '0-0 2-1 4-2 6-3 6-4 8-5 10-6 10-7 12-9 13-11',
        ])
        mod = RemoveEndPunctuationModifier(1.0)
        out = list(mod([ex]))[0]
        self.assertEqual(out, ex)

    def test_do_not_remove_ascii_ellipsis(self):
        ex = '\t'.join([
            'This ▁ is ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣 ▁ . . .',
            '这 是 一个 简单 的 测试 语 句 ▁ 🤣 ▁ .',
            '0-0 2-1 4-2 6-3 6-4 8-5 10-6 10-7 12-9 13-11 14-11 15-11',
        ])
        mod = RemoveEndPunctuationModifier(1.0)
        out = list(mod([ex]))[0]
        self.assertEqual(out, ex)

    def test_do_not_remove_ascii_ellipsis_untokenized(self):
        ex = '\t'.join([
            'This is a simple test statement🤣...',
            '这是一个简单的测试语句🤣.',
        ])
        mod = RemoveEndPunctuationModifier(1.0)
        out = list(mod([ex]))[0]
        self.assertEqual(out, ex)



from typing import Optional, List, Union
from opustrainer.types import Pair, TokenList


def parse_alignments(input_pairs:str, src_tokens:Optional[TokenList]=None, trg_tokens:Optional[TokenList]=None) -> List[Pair]:
    """Parses `1-2 3-4` into `[Pair(src=1,trg=2), Pair(src=3,trg=4)]`. If `src_tokens` and
    `trg_tokens` are supplied, it will also check that the indices are not out of bounds."""
    pairs: List[Pair] = [
        Pair(int(a), int(b)) for a, b in
        (pair.split('-', maxsplit=1) for pair in input_pairs.split())
    ]

    if src_tokens is not None and trg_tokens is not None:
        for pair in pairs:
            if pair.src < 0 or pair.src >= len(src_tokens) \
            or pair.trg < 0 or pair.trg >= len(trg_tokens):
                raise ValueError('Out-of-bound alignment pairs')

    return pairs


def format_alignments(pairs:List[Pair]) -> str:
    """Opposite of `parse_alignments`, turns a list of alignments back into the `a-b c-d ...` string
    format that most alignment tools expect."""
    return ' '.join(f'{pair.src}-{pair.trg}' for pair in pairs)



from typing import List, Dict, NamedTuple, Tuple, Iterable

from opustrainer.types import Pair, TokenList, TokenMapping, Tokenizer, Detokenizer
from opustrainer.alignments import parse_alignments, format_alignments
from opustrainer.tokenizers import make_tokenizer, make_detokenizer
from opustrainer.modifiers import Modifier
from opustrainer import logger


def slice_cmp(r1:slice, r2:slice) -> int:
    """Compare how two slices relate to each other.
       Returns -1 if r1 is before r2,
                0 if r1 and r2 overlap
                1 if r1 is after r2
    """
    if r1.stop <= r2.start:
        return -1
    elif r1.start >= r2.stop:
        return 1
    else:
        return 0


class Retokenizer(NamedTuple):
    detokenizer: Detokenizer
    tokenizer: Tokenizer

    def retokenize(self, tokens:TokenList) -> Tuple[str,TokenList,TokenMapping]:
        detokenized, old_token_spans = self.detokenizer.detokenize(tokens)
        new_tokens, new_token_spans = self.tokenizer.tokenize(detokenized)

        old_to_new_mapping = [[] for _ in range(len(old_token_spans))]

        prev_j = 0
        for i, old_token_span in enumerate(old_token_spans):
            # it is possible for ICU tokenizer whitespace token, return empty list
            if old_token_span is None:
                continue

            for j, new_token_span in enumerate(new_token_spans[prev_j:], start=prev_j):
                prev_j = j
                overlap = slice_cmp(old_token_span, new_token_span)
                if overlap < 0: # old_token_span is before new_token_span
                    break # skip to next old_token_span
                elif overlap > 0: # old_token_span is after new_token_span
                    continue # skip forward to next new_token_span
                else:
                    old_to_new_mapping[i].append(j) # overlap!

        return detokenized, new_tokens, old_to_new_mapping


def make_retokenizer(spec:Dict[str,str]) -> Retokenizer:
    return Retokenizer(
        detokenizer=make_detokenizer(spec.get('detokenize', 'spaces')),
        tokenizer=make_tokenizer(spec.get('tokenize', 'spaces'))
    )


def remap_alignment_pairs(src_mapping:TokenMapping, trg_mapping:TokenMapping, alignments:List[Pair]) -> List[Pair]:
    """Will recalculate the alignment pairs to match a new tokenization scheme
    according to the updated mappings for the source and target side of the
    sentence pair.
    
    E.g. if you have
    source-mapping: [0 => [3,4], 1 => [5], 2 => []],
    target-mapping: [0 => [0], 1 => [1], 2 => []]
    alignments:     [(0,1), (1,1)]
    it will return  [
        (3,1), (4,1), # the [0 => [3,4]] mapping
        (5,1)         # the [1 => [5]] mapping
    ]
    """
    remapped = set()
    for old_src_idx, old_trg_idx in alignments:
        for src_idx in src_mapping[old_src_idx]:
            for trg_idx in trg_mapping[old_trg_idx]:
                remapped.add(Pair(src_idx, trg_idx))
    return sorted(remapped)


class RetokenizeModifier(Modifier):
    """Retokenizes the input line, fixing up the alignments *but giving you the detokenized text*.
    The probability argument is ignored. Most of this functionality is already built into the Tags
    placeholder.

    Default detokenizer and tokenizer are `spaces`, which just splits and joins the tokens with a
    space in between. Other options are `moses:{lang}` for Moses detokenizer and tokenizer, and
    `spm` for sentencepiece.

    `spm` is only available as a tokenize option at this moment since other modifiers won't know
    how to deal with spm input anyway.
    
    Usage:
    ```yaml
    modifiers:
    - Retokenize: 0
      src:
        detokenize: moses:en
        tokenize: spm:path/to/vocab.spm
      trg:
        detokenize: moses:zh
        tokenize: spm:path/to/vocab.spm

    ```
    """

    src: Retokenizer
    trg: Retokenizer

    def __init__(self, probability: float=0.0, src:dict=dict(), trg:dict=dict()):
        super().__init__(probability) # probability is very much ignored lol.
        self.src = make_retokenizer(src)
        self.trg = make_retokenizer(trg)

    def __call__(self, batch:List[str]) -> Iterable[str]:
        for line in batch:
            try:
                src, trg, alignments = line.split('\t')
                src_tokens = src.split()
                trg_tokens = trg.split()
                pairs = parse_alignments(alignments, src_tokens, trg_tokens)
                new_src, _, src_mapping = self.src.retokenize(src_tokens)
                new_trg, _, trg_mapping = self.trg.retokenize(trg_tokens)
                remapped_pairs = remap_alignment_pairs(src_mapping, trg_mapping, pairs)
                yield '\t'.join((new_src, new_trg, format_alignments(remapped_pairs)))
            except Exception as exc:
                logger.log(f'Exception while processing line, skipping line: {exc!r}', 'WARNING')



import re
from typing import Tuple, List, TypeVar, Callable, Union, Dict, Optional
from pathlib import Path

import sacremoses
from sentencepiece import SentencePieceProcessor

from opustrainer.types import TokenList, TokenSpanList, Tokenizer, Detokenizer


DETOKENIZERS = {
    'moses': lambda lang: MosesDetokenizer(lang),
    'spaces': lambda: SpaceDetokenizer(),
    'icu': lambda lang: IcuDetokenizer(lang),

}

TOKENIZERS = {
    'moses': lambda lang: MosesTokenizer(lang),
    'spm': lambda vocab: SentencePieceTokenizer(vocab),
    'spaces': lambda: SpaceTokenizer(),
    'icu': lambda lang: IcuTokenizer(lang),
}


T = TypeVar('T', bound=Union[Tokenizer,Detokenizer])

def _make(implementations: Dict[str,Callable[...,T]], spec:str) -> T:
    name, *args = spec.split(':')
    return implementations[name](*args)

def make_detokenizer(spec:str) -> Detokenizer:
    """Creates a Detokenizer using a spec, e.g. `moses:en` or spm:path/to/vocab.spm`."""
    return _make(DETOKENIZERS, spec)

def make_tokenizer(spec:str) -> Tokenizer:
    """Creates a Tokenizer using a spec, e.g. `moses:en` or spm:path/to/vocab.spm`."""
    return _make(TOKENIZERS, spec)


class SpaceTokenizer:
    """Splits 'Hello World.' into `[Hello, World.]`."""
    def tokenize(self, text:str) -> Tuple[TokenList, TokenSpanList]:
        tokens: TokenList = []
        spans: TokenSpanList = []
        for match in re.finditer(r'[^\s]+', text):
            tokens.append(match.group(0))
            spans.append(slice(match.start(0), match.end(0)))
        return tokens, spans


class SpaceDetokenizer:
    """Turns `[Hello, World.]` back into `Hello World.`."""
    def detokenize(self, tokens:TokenList) -> Tuple[str,TokenSpanList]:
        spans = []
        offset = 0
        for token in tokens:
            spans.append(slice(offset, offset + len(token)))
            offset += len(token) + 1 # space
        return ' '.join(tokens), spans


class MosesTokenizer:
    """Turns `Hello World.` into `[Hello, World, .]` according to Moses and,
    if available, language specific rules."""
    tokenizer: sacremoses.MosesTokenizer

    def __init__(self, lang:str, custom_nonbreaking_prefixes:Optional[str]=None):
        self.tokenizer = sacremoses.MosesTokenizer(lang, custom_nonbreaking_prefixes)

    def tokenize(self, text:str) -> Tuple[TokenList, TokenSpanList]:
        tokens:TokenList = self.tokenizer.tokenize(text, escape=False) # type: ignore
        spans: TokenSpanList = [] # ^tokenizer.tokenize always returns a string unless return_string=True (which is not)
        offset = 0
        for token in tokens:
            offset = text.find(token, offset)
            if offset == -1:
                raise RuntimeError(f"Could not find token '{token}' in original text")
            spans.append(slice(offset, offset + len(token)))
            offset += len(token)
        return tokens, spans


class MosesDetokenizer:
    """Turns `[Hello,World,.]` back into `Hello World.`. Rules can be language-specific."""
    detokenizer: sacremoses.MosesDetokenizer

    def __init__(self, lang:str):
        self.detokenizer = sacremoses.MosesDetokenizer(lang)

    def detokenize(self, tokens:TokenList) -> Tuple[str,TokenSpanList]:
        text = self.detokenizer.detokenize(tokens)
        spans = []
        offset = 0
        for token in tokens:
            offset = text.find(token, offset)
            if offset == -1:
                raise RuntimeError(f"Could not find token '{token}' in detokenized text")
            spans.append(slice(offset, offset + len(token)))
            offset += len(token)
        return text, spans


class SentencePieceTokenizer:
    """Turns `Hello World.` into something like [He,llo,_World,.] depending on your vocab."""
    spm: SentencePieceProcessor

    def __init__(self, vocab:Path):
        self.spm = SentencePieceProcessor(model_file=str(vocab)) # type: ignore # for some reason pylance doesn't understand spm.

    def tokenize(self, text:str) -> Tuple[TokenList,TokenSpanList]:
        # interestingly, piece.begin and piece.end are unicode offsets, not byte
        # offsets as the documentation would suggest. When byte-fallback happens,
        # there will be pieces where piece.begin and piece.end are the same value
        # but they are technically necessary to encode the following pieces.
        # e.g:
        # > x.encode('🤣', out_type='immutable_proto').pieces
        #   { piece: "▁" id: 275 surface: "" begin: 0 end: 0 }
        #   { piece: "<0xF0>" id: 247 surface: "" begin: 0 end: 0 }
        #   { piece: "<0x9F>" id: 166 surface: "" begin: 0 end: 0 }
        #   { piece: "<0xA4>" id: 171 surface: "" begin: 0 end: 0 }
        #   { piece: "<0xA3>" id: 170 surface: "🤣" begin: 0 end: 1 }
        # > x.decode([247,166,171,170])
        #   '🤣'
        spans = [
            slice(piece.begin, piece.end)
            for piece in self.spm.encode(text.encode(), out_type='immutable_proto').pieces # type: ignore
        ]
        tokens = [text[span] for span in spans]
        return tokens, spans

# The same character as in SentencePiece
ICU_WHITESPACE_TOKEN = "▁"
class IcuTokenizer:
    """
    Tokenizes text by splitting words and punctuation using ICU segmenter.
    Whitespaces will be preserved as a special token ▁ for lossless detokenization.
    Requires installation with the steps specified in https://pypi.org/project/PyICU/
    """

    def __init__(self, lang: str):
        self.lang = lang

    def tokenize(self, text:str) -> Tuple[TokenList, TokenSpanList]:
        from icu import BreakIterator, Locale

        bi = BreakIterator.createWordInstance(Locale(self.lang))
        bi.setText(text)

        tokens = []
        start = bi.first()
        for end in bi:
            token = text[start:end]
            if (
                token and token != "\n"
            ):  # exclude empty tokens, but leave whitespaces and replace them with a special token
                tokens.append(token)
            start = end

        spans: TokenSpanList = []
        offset = 0
        for token in tokens:
            offset = text.find(token, offset)
            if offset == -1:
                raise RuntimeError(f"Could not find token '{token}' in original text")
            spans.append(slice(offset, offset + len(token)))
            offset += len(token)

        tokens = [token.replace(" ", ICU_WHITESPACE_TOKEN) for token in tokens]
        return tokens, spans

class IcuDetokenizer:
    """
    Detokenizes tokens back into the original text preserving whitespaces as well.
    Spans for whitespaces will be None.
    """

    # For compatibility with MosesDetokenizer interface
    def __init__(self, lang):
        self.lang = lang

    def detokenize(self, tokens:TokenList) -> Tuple[str,TokenSpanList]:
        text = "".join(tokens).replace(ICU_WHITESPACE_TOKEN, " ")

        spans = []
        offset = 0

        for token in tokens:
            if token == ICU_WHITESPACE_TOKEN:
                spans.append(None)
                continue
            # there are some edge cases where a whitespace can appear inside a token
            token = token.replace(ICU_WHITESPACE_TOKEN, " ")
            offset = text.find(token, offset)
            if offset == -1:
                raise RuntimeError(f"Could not find token '{token}' in detokenized text")
            spans.append(slice(offset, offset + len(token)))
            offset += len(token)

        return text, spans



import random
from typing import List, Iterable
from opustrainer.modifiers import Modifier


class PrefixModifier(Modifier):
    """Prefixes the source sentence with a phrase from the target sentence. 
    Note that if the target is ZH, JA or basically any other language that is not space segmented,
    this wouldn't work as we would be prefixing the whole sentence (or very large part of it).
    This is not supported for now. A proper fix would be to use custom detokenizer like in the placeholders.

       Usage:
       ```yaml
       modifiers:
       - Prefix: 0.02
         min_words: 2
         max_words: 5
         template: "__start__ {trg} __end__ "
        ```
    """
    min_words: int
    max_words: int
    template: str

    def __init__(self, probability: float=0.0, min_words: int=2,
        max_words: int=5, template: str="__start__ {trg} __end__ "):
        super().__init__(probability)
        self.min_words = min_words
        self.max_words = max_words
        self.template = template

    def __call__(self, batch:List[str]) -> Iterable[str]:
        for line in batch:
            yield self.apply(line)

    def apply(self, line:str) -> str:
        """Takes a line in the form of "I like pie. Me gustan los pasteles."
           and turns it into: "__start__ los pasteles __end__ I like pie. Me
           gustan los pasteles."
        """
        if self.probability < random.random():
            return line
        
        # Take care of cases where we could have the alignment info at the end.
        fields: List[str] = line.split('\t')
        target_tok: List[str] = fields[1].split()

        # determine the length of the sample
        num_tokens = random.randint(self.min_words, self.max_words)

        # Select start token, accounting for the fact that we need at least num_tokens in the sentence
        max_start_token = len(target_tok) - num_tokens

        # If we have a negative max_start_token it means that the sentence
        # is too short for this augmentation and we should just skip it.
        if max_start_token < 0:
            return line

        # random.randrange(x) generates a number in the interval of [0, X)
        # max_start_token is computed as the difference in length of the two sequences which means that
        # we want random.ranrange(x) to produce [0, X], therefore we increment it by one.
        start_token = random.randrange(max_start_token + 1)
        
        augment_substring:str  = " ".join(target_tok[start_token:start_token + num_tokens])

        # Return the modified string
        return self.template.format(trg=augment_substring) + line



import random
from typing import List, Iterable

from opustrainer.modifiers import Modifier


class TitleCaseModifier(Modifier):
    """Applies titlecase to a sentence. Beware of tabs as src and trg separator
    """
    def __call__(self, batch:List[str]) -> Iterable[str]:
        for line in batch:
            if self.probability <= random.random():
                yield line
            else:
                sections: List[str] = line.split('\t')
                for i in range(len(sections)):
                    sections[i] = ' '.join([word[0].upper() + word[1:] for word in sections[i].split()])
                yield '\t'.join(sections)


class UpperCaseModifier(Modifier):
    def __call__(self, batch:List[str]) -> Iterable[str]:
        for line in batch:
            yield line.upper() if self.probability > random.random() else line



import random
from operator import attrgetter
from pathlib import Path
from typing import Set, List, Tuple, Optional, TypeVar, Iterable

from opustrainer.alignments import Pair, parse_alignments, format_alignments
from opustrainer.modifiers import Modifier
from opustrainer.tokenizers import SpaceDetokenizer, SpaceTokenizer, SentencePieceTokenizer, \
    make_detokenizer, ICU_WHITESPACE_TOKEN
from opustrainer.modifiers.retokenize import Retokenizer, remap_alignment_pairs
from opustrainer import logger


T = TypeVar('T')

def random_weighted_choice(options:Iterable[Tuple[T,float]]) -> T:
    choice = random.random()
    cumsum = 0.0
    for option, prob in options:
        cumsum += prob
        if choice < cumsum:
            return option
    raise RuntimeError('random_weighted_choice called with cumulative sum smaller than 1.0')


def first(iterable:Iterable[T]) -> T:
    """Returns the first value of an iterable. Might raise a StopIteration if iterable is empty"""
    return next(iter(iterable))


def get_random_unicode_words(min_length: int=2, max_length: int=10, max_words: int=3) -> List[str]:
    """Gets a random unicode string of words, of up to max_words, where each word is of length
    min_length-max_length. Only one character set per invocation.
    Maybe should do special rules for emoji and CJK? Emoji wouldn't appear with spaces in between normally
    and CJK would normally be of maximum length of 3 per word."""

    length: int = random.randint(min_length, max_length)

    # Update this to include code point ranges to be sampled
    # https://jrgraphix.net/r/Unicode/
    # https://en.wikipedia.org/wiki/Unicode_block
    include_ranges = [
        (0x0020, 0x007F), # Basic Latin
        (0x00A0, 0x00FF), # Latin-1 Supplement
       # (0x0100, 0x017F), # Latin Extended-A
       # (0x0180, 0x024F), # Latin Extended-B
       # (0x0250, 0x02AF), # IPA Extensions
       # (0x02B0, 0x02FF), # Spacing Modifier Letters
       # (0x0300, 0x036F), # Combining Diacritical Marks
        (0x0370, 0x03FF), # Greek and Coptic
        (0x0400, 0x04FF), # Cyrillic
       # (0x0500, 0x052F), # Cyrillic Supplementary
        (0x0530, 0x058F), # Armenian
        (0x0590, 0x05FF), # Hebrew
        (0x0600, 0x06FF), # Arabic
       # (0x0700, 0x074F), # Syriac
       # (0x0780, 0x07BF), # Thaana
        (0x0900, 0x097F), # Devanagari
        (0x0980, 0x09FF), # Bengali
       # (0x0A00, 0x0A7F), # Gurmukhi
        (0x0A80, 0x0AFF), # Gujarati
       # (0x0B00, 0x0B7F), # Oriya
       # (0x0B80, 0x0BFF), # Tamil
       # (0x0C00, 0x0C7F), # Telugu
       # (0x0C80, 0x0CFF), # Kannada
       # (0x0D00, 0x0D7F), # Malayalam
       # (0x0D80, 0x0DFF), # Sinhala
        (0x0E00, 0x0E7F), # Thai
       # (0x0E80, 0x0EFF), # Lao
       # (0x0F00, 0x0FFF), # Tibetan
        (0x1000, 0x109F), # Myanmar
        (0x10A0, 0x10FF), # Georgian
       # (0x1100, 0x11FF), # Hangul Jamo
       # (0x1200, 0x137F), # Ethiopic
       # (0x13A0, 0x13FF), # Cherokee
       # (0x1400, 0x167F), # Unified Canadian Aboriginal Syllabics
       # (0x1680, 0x169F), # Ogham
       # (0x16A0, 0x16FF), # Runic
       # (0x1700, 0x171F), # Tagalog
       # (0x1720, 0x173F), # Hanunoo
       # (0x1740, 0x175F), # Buhid
       # (0x1760, 0x177F), # Tagbanwa
        (0x1780, 0x17FF), # Khmer
       # (0x1800, 0x18AF), # Mongolian
       # (0x1900, 0x194F), # Limbu
       # (0x1950, 0x197F), # Tai Le
       # (0x19E0, 0x19FF), # Khmer Symbols
       # (0x1D00, 0x1D7F), # Phonetic Extensions
       # (0x1E00, 0x1EFF), # Latin Extended Additional
       # (0x1F00, 0x1FFF), # Greek Extended
       # (0x2000, 0x206F), # General Punctuation
       # (0x2070, 0x209F), # Superscripts and Subscripts
       # (0x20A0, 0x20CF), # Currency Symbols
       # (0x20D0, 0x20FF), # Combining Diacritical Marks for Symbols
       # (0x2100, 0x214F), # Letterlike Symbols
       # (0x2150, 0x218F), # Number Forms
       # (0x2190, 0x21FF), # Arrows
       # (0x2200, 0x22FF), # Mathematical Operators
       # (0x2300, 0x23FF), # Miscellaneous Technical
       # (0x2400, 0x243F), # Control Pictures
       # (0x2440, 0x245F), # Optical Character Recognition
       # (0x2460, 0x24FF), # Enclosed Alphanumerics
       # (0x2500, 0x257F), # Box Drawing
       # (0x2580, 0x259F), # Block Elements
       # (0x25A0, 0x25FF), # Geometric Shapes
       # (0x2600, 0x26FF), # Miscellaneous Symbols
       # (0x2700, 0x27BF), # Dingbats
       # (0x27C0, 0x27EF), # Miscellaneous Mathematical Symbols-A
       # (0x27F0, 0x27FF), # Supplemental Arrows-A
       # (0x2800, 0x28FF), # Braille Patterns
       # (0x2900, 0x297F), # Supplemental Arrows-B
       # (0x2980, 0x29FF), # Miscellaneous Mathematical Symbols-B
       # (0x2A00, 0x2AFF), # Supplemental Mathematical Operators
       # (0x2B00, 0x2BFF), # Miscellaneous Symbols and Arrows
       # (0x2E80, 0x2EFF), # CJK Radicals Supplement
       # (0x2F00, 0x2FDF), # Kangxi Radicals
       # (0x2FF0, 0x2FFF), # Ideographic Description Characters
       # (0x3000, 0x303F), # CJK Symbols and Punctuation
        (0x3040, 0x309F), # Hiragana
        (0x30A0, 0x30FF), # Katakana
       # (0x3100, 0x312F), # Bopomofo
       # (0x3130, 0x318F), # Hangul Compatibility Jamo
       # (0x3190, 0x319F), # Kanbun
       # (0x31A0, 0x31BF), # Bopomofo Extended
       # (0x31F0, 0x31FF), # Katakana Phonetic Extensions
       # (0x3200, 0x32FF), # Enclosed CJK Letters and Months
       # (0x3300, 0x33FF), # CJK Compatibility
       # (0x3400, 0x4DBF), # CJK Unified Ideographs Extension A
       # (0x4DC0, 0x4DFF), # Yijing Hexagram Symbols
        (0x4E00, 0x9FFF), # CJK Unified Ideographs
       # (0xA000, 0xA48F), # Yi Syllables
       # (0xA490, 0xA4CF), # Yi Radicals
        (0xAC00, 0xD7AF), # Hangul Syllables
       # (0xD800, 0xDB7F), # High Surrogates
       # (0xDB80, 0xDBFF), # High Private Use Surrogates
       # (0xDC00, 0xDFFF), # Low Surrogates
       # (0xE000, 0xF8FF), # Private Use Area
       # (0xF900, 0xFAFF), # CJK Compatibility Ideographs
       # (0xFB00, 0xFB4F), # Alphabetic Presentation Forms
       # (0xFB50, 0xFDFF), # Arabic Presentation Forms-A
       # (0xFE00, 0xFE0F), # Variation Selectors
       # (0xFE20, 0xFE2F), # Combining Half Marks
       # (0xFE30, 0xFE4F), # CJK Compatibility Forms
       # (0xFE50, 0xFE6F), # Small Form Variants
       # (0xFE70, 0xFEFF), # Arabic Presentation Forms-B
       # (0xFF00, 0xFFEF), # Halfwidth and Fullwidth Forms
       # (0xFFF0, 0xFFFF), # Specials
       # (0x10000, 0x1007F), # Linear B Syllabary
       # (0x10080, 0x100FF), # Linear B Ideograms
       # (0x10100, 0x1013F), # Aegean Numbers
       # (0x10300, 0x1032F), # Old Italic
       # (0x10330, 0x1034F), # Gothic
       # (0x10380, 0x1039F), # Ugaritic
       # (0x10400, 0x1044F), # Deseret
       # (0x10450, 0x1047F), # Shavian
       # (0x10480, 0x104AF), # Osmanya
       # (0x10800, 0x1083F), # Cypriot Syllabary
       # (0x1D000, 0x1D0FF), # Byzantine Musical Symbols
       # (0x1D100, 0x1D1FF), # Musical Symbols
       # (0x1D300, 0x1D35F), # Tai Xuan Jing Symbols
       # (0x1D400, 0x1D7FF), # Mathematical Alphanumeric Symbols
        (0x1F600, 0x1F64F),  # Emoji
       # (0x20000, 0x2A6DF), # CJK Unified Ideographs Extension B
       # (0x2F800, 0x2FA1F), # CJK Compatibility Ideographs Supplement
       # (0xE0000, 0xE007F), # Tags
    ]
    # Select a character set
    alphabet = random.choice(include_ranges)
    
    # Generate a random string of 1 - 3 words
    return [
        ''.join(chr(random.randrange(*alphabet)) for _ in range(length))
        for _ in range(random.randint(1, max_words))
    ]


def filter_one_to_one_pairs(inlist: List[Pair]) -> List[Pair]:
    """Removes places non x->y x->z type of alignments. Remove anything that is found multiple
       times. Anything that is found multiple times means non bijective alignment. Since they
       are sorted, we can reduce some time complexity
    """
    inlist.sort(key=attrgetter('src'))
    new_list: List[Pair] = []
    blacklisted: Set[int] = set()
    for i, pair in enumerate(inlist):
        # Skip blacklisted ones
        if pair.src in blacklisted:
            continue
        
        # Skip (and blacklist) any (x,y) pairs that are followed by (x,_) pairs
        if i < len(inlist) - 1 and pair.src == inlist[i+1].src:
            blacklisted.add(pair.src)
            continue

        # Keep the rest, pairs like (x,_) where `x` did only occur once in the
        # first position among all of the tuples in `inlist`.
        new_list.append(pair)
    return new_list


def get_placeholding_candidates(src_trg: List[Pair]) -> List[Pair]:
    """Filters out multiple alignment targets so that we can definitely get a one-to-one
       replacement
    """
    # Create the two src-trg and trg-src sets
    trg_src: List[Pair] = [Pair(pair.trg, pair.src) for pair in src_trg]

    src_trg_filtered: List[Pair] = filter_one_to_one_pairs(src_trg)
    trg_src_filtered: List[Pair] = filter_one_to_one_pairs(trg_src)

    # Now, we are looking for the union of the two.
    # First, reverse src_trg
    trg_src_filtered_rereversed: List[Pair] = [Pair(pair.trg, pair.src) for pair in trg_src_filtered]

    # Now convert both to sets and take the union
    src_trg_set: Set[Pair] = set(src_trg_filtered)
    trg_src_filtered_rereversed_set: Set[Pair] = set(trg_src_filtered_rereversed)
    selection = src_trg_set & trg_src_filtered_rereversed_set

    # Little dance to make sure we return the original pair instances, not any
    # of the reversed reversed ones. This also sorts them back into the original
    # order.
    return [pair for pair in src_trg if pair in selection]


class PlaceholderTagModifier(Modifier):
    """Unpacks a line, removes the alignments, and applies placeholding. Supports trivial and non trivial detokenization
       using moses detokenizer, which should be used for CJK languages and languages where words are typically not space
       delimited.

       Usage:
       ```yaml
       modifiers:
       - Tags: 0.02
         custom_detok_src: 'moses:zh'
         custom_detok_trg: "moses:null"
         template: "__source__ {src} __target__ {trg} __done__"
         augment: 0.0 # 0% chance to just insert a random string on both sides
         replace: 0.0 # 0% change to use tags to force translate to a random string
        ```
    """

    template: str

    src_retokenizer: Retokenizer
    trg_retokenizer: Retokenizer

    modes: List[Tuple[str,float]]

    # Controls whether alignment info is printed. Normally this is controlled by
    # `spm_vocab` argument, but in tests we also set this to True for testing.
    print_alignments: bool

    def __init__(self, probability: float=0.0, custom_detok_src: Optional[str]=None, custom_detok_trg: Optional[str]=None,
        spm_vocab_src: Optional[Path]=None, spm_vocab_trg: Optional[Path]=None,
        template: str="__source__ {src} __target__ {trg} __done__", augment: float=0, replace:float=0, tag:float=1):
        super().__init__(probability)

        self.template = template

        # uses Moses detokenizer by default
        if custom_detok_src and ':' not in custom_detok_src:
            custom_detok_src = f'moses:{custom_detok_src}'
        if custom_detok_trg and ':' not in custom_detok_trg:
            custom_detok_trg = f'moses:{custom_detok_trg}'

        self.custom_detok_src = custom_detok_src
        self.custom_detok_trg = custom_detok_trg

        self.src_retokenizer = Retokenizer(
            detokenizer=make_detokenizer(custom_detok_src) if custom_detok_src else SpaceDetokenizer(),
            tokenizer=SentencePieceTokenizer(spm_vocab_src) if spm_vocab_src else SpaceTokenizer()
        )

        self.trg_retokenizer = Retokenizer(
            detokenizer=make_detokenizer(custom_detok_trg) if custom_detok_trg else SpaceDetokenizer(),
            tokenizer=SentencePieceTokenizer(spm_vocab_trg) if spm_vocab_trg else SpaceTokenizer()
        )

        # For now only print alignments when spm_vocab is passed in. We'll improve upon this with #29.
        self.print_alignments = spm_vocab_trg is not None

        self.modes = []

        if augment + replace > 1.0:
            raise ValueError('sum of augment and replace probability should not exceed 1.0')

        if augment > 0:
            self.modes.append(('augment', augment))

        if replace > 0:
            self.modes.append(('replace', replace))

        # the modifier can be used for inline noise augmentation only
        if tag > 0:
            self.modes.append(('tag', tag))

        if ({'replace', 'tag'} & {mode for mode,_ in self.modes}) and \
            'icu' in ((self.custom_detok_trg or '') + (self.custom_detok_trg or '')):
            raise ValueError('ICU tokenization is not supported with "tag" and "replace" modes')

    def __call__(self, batch: List[str]) -> Iterable[str]:
        for line in batch:
            try:
                yield self.apply(line)
            except Exception as exc:
                logger.log(f'Skipping line because of exception: {exc!r}', 'WARNING')

    def apply(self, line:str) -> str:
        """Applies tag to words in a line based on alignment info, and then removes the alignment info from the line.
           This is used to enable terminology support by tagging random words with their translation.
           eg "I like cake" would become "I __source__ like __target__ gusta __done__ cake.
           By default the detokenizer used is the trivial detokenizer, but we can instead have separate detokenizers on src and trg."
        """

        src, trg, *rest = line.strip().split('\t')
        source = src.split()
        target = trg.split()
        alignments = []
        
        # Try parsing alignments. If we fail, the sentence will be thrown out
        # by the trainer.
        alignments = parse_alignments(rest[0], source, target)
        candidate_offset = 0;

        while self.probability > 0.0:
            try:
                # Get replacement candidate. Skip words that are already the same in the source and
                # target sentence. This is to avoid having numbers trained with placeholders or any
                # other words that are exactly the same.
                candidate = first(
                    candidate
                    for candidate in get_placeholding_candidates(alignments[candidate_offset:])
                    if source[candidate.src] != target[candidate.trg]
                )
            except StopIteration:
                # `first` failed because there are no sutable candidates, so lets break out of this
                # loop as well and skip to the output printing part.
                break

            # Candidate pair position in the alignments array
            candidate_index = alignments.index(candidate)

            # Default for if we skip this word: next candidate is found after the current one
            candidate_offset = candidate_index + 1

             # Skip words whose turn it isn't yet.
            if random.random() >= self.probability:
                continue
            
            # Select mode (skip random_weighted_choices*() when 'tag' is the only mode)
            mode = random_weighted_choice(self.modes) if len(self.modes) > 1 else self.modes[0][0]

            if mode == "tag" or mode == "replace":
                if mode == "tag":
                    # Hint the expected output to the trainer by specifying it in the input as a tag
                    augment_tokens = [target[candidate.trg]]
                else:
                    # Same as above, but instead of the expected target word, we replace it on both
                    # sides with a random string. This encourages the model to really ignore the source
                    # and produce the target that we desire.
                    augment_tokens = get_random_unicode_words()
                
                tag_tokens = self.template.format(src=source[candidate.src], trg=' '.join(augment_tokens)).split()
                source = source[:candidate.src] + tag_tokens + source[candidate.src+1:]
                target = target[:candidate.trg] + augment_tokens + target[candidate.trg+1:]

                src_tag_offset = first(n for n, tpl in enumerate(self.template.split()) if '{src}' in tpl)
                trg_tag_offset = first(n for n, tpl in enumerate(self.template.split()) if '{trg}' in tpl)

                # __src__ aaa __trg__ xxx yyy zzz __done__ => xxx yyy zzz
                # ^1      ^2  ^3      ^4  ^5  ^6  ^7          ^1  ^2  ^3
                # So what is the correct alignment pairs for these? For now
                # I map 2-1 2-2 2-3 + 4-1 5-2 6-3?

                # Fix up alignment pairs
                alignments = (
                    # pairs before the replaced bit stay the same
                    alignments[:candidate_index]
                    # fill in the gap created by the replaced bit: the 2-1 2-2 2-3 bit
                    + [Pair(candidate.src + src_tag_offset, candidate.trg + n) for n in range(len(augment_tokens))]
                    # fill in the gap created by the replaced bit: the 4-1 5-2 6-3 bit
                    + [Pair(candidate.src + trg_tag_offset + n, candidate.trg + n) for n in range(len(augment_tokens))]
                    # pairs after the replaced bit have to be offset by the length of the replacement
                    # bit minus the length of the bit we replaced (1)
                    + [Pair(pair.src + len(tag_tokens) - 1, pair.trg) for pair in alignments[candidate_index+1:]]
                )
                candidate_offset = candidate_index + 2 * len(augment_tokens)

            elif mode == "augment":
                # Augment mode adds random noise both on the source and the target without any
                # tagging encouraging the model to copy crap from one side to the other.
                augment_tokens = get_random_unicode_words()
                source, num_src_aug_tokens, pos_aug_src = self.insert_augmented(augment_tokens, source, candidate.src+1, self.custom_detok_src)
                target, num_trg_aug_tokens, pos_aug_trg = self.insert_augmented(augment_tokens, target, candidate.trg+1, self.custom_detok_trg)

                # Fix up alignment pairs
                alignments = (
                    # pairs before and including the candidate stay the same
                    alignments[:candidate_index+1]
                    # fill in the gap created by the added random noise
                    + [Pair(candidate.src + n_src, candidate.trg + n_trg) for n_src, n_trg in zip(pos_aug_src, pos_aug_trg)]
                    # pairs after the replaced bit have to be offset by the length of the replacement bit
                    + [Pair(pair.src + num_src_aug_tokens, pair.trg + num_trg_aug_tokens) for pair in alignments[candidate_index+1:]]
                )
                candidate_offset = candidate_index + min(num_src_aug_tokens, num_trg_aug_tokens) + 1

        source_detok, _, source_mapping = self.src_retokenizer.retokenize(source)
        target_detok, _, target_mapping = self.trg_retokenizer.retokenize(target)
        
        if self.print_alignments:
            remapped_pairs = remap_alignment_pairs(source_mapping, target_mapping, alignments)
            return source_detok + "\t" + target_detok + "\t" + format_alignments(remapped_pairs)
        else:
            return source_detok + "\t" + target_detok

    def insert_augmented(self, augment_tokens: List[str], tokens: List[str], position: int, detokenization: str) -> Tuple[List[str], int, List[int]]:
        """
        Inserts augmented tokens.
        Accounts for possible ICU detokenization which uses special symbol "▁" for whitespace tokens.
            Such tokens will also be inserted to separate the augmented words.

        Returns:
            new tokens
            number of augmented tokens including whitespaces for in ICU case
            alignments positions for the augmented tokens (whitespaces are excluded, we don't need alignments for them)
        """
        prefix = tokens[:position]
        postfix = tokens[position:]
        aug_aln_offset = []

        if detokenization is not None and "icu" in detokenization:
            new_aug_tokens = []
            aug_pos_index = 1

            if len(prefix) > 0 and prefix[-1] != ICU_WHITESPACE_TOKEN:
                new_aug_tokens.append(ICU_WHITESPACE_TOKEN)
                aug_pos_index += 1

            for token in augment_tokens:
                new_aug_tokens.append(token)
                # save the offset of the augmented words to use in alignments
                aug_aln_offset.append(aug_pos_index)
                new_aug_tokens.append(ICU_WHITESPACE_TOKEN)
                aug_pos_index += 2

            if len(postfix) > 0 and postfix[0] == ICU_WHITESPACE_TOKEN:
                new_aug_tokens.pop()

            augment_tokens = new_aug_tokens
        else:
            aug_aln_offset = list(range(1, len(augment_tokens) + 1))

        tokens = prefix + augment_tokens + postfix
        return tokens, len(augment_tokens), aug_aln_offset

    def validate(self, context:List[Modifier]) -> None:
        """Current limitation of the tags modifier is that any other modifier might modify the
        inserted tags, which we don't want. So warn users about that if we notice it.
        """
        if context[-1] != self:
            logger.log('Tags modifier should to be the last modifier to be applied, as otherwise other modifiers might alter the inserted tags themselves.', loglevel="WARNING")
